/* $Id: espda_convol_synth.c 2018-10-18 14:15:10 jonaydas $
 *
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Author: jonaydas $
 * $Date: 2018-10-18 14:15:10 +0000 (sat, 24 feb 2017) $
 * $Revision: 202524 $
 * $Name:  $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>

#include "espda_convol_synth.h"

#include "espda_dfs.h"
#include "espda_frame.h"
#include "espda_spec.h"

/**@{*/

/*----------------------------------------------------------------------------*/
/**
 * @defgroup espda_convol_synth
 *
 * Functions for convolution of synthetic spectrum
 */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Convolve synthetic spectrum with broadening profiles
  @param    wobs     Array of wavelength of the observed spectrum
  @param    fsyn     Array of flux of the interpolated synthetic spectrum
  @param    fcon     Array of flux of the convolved, interpolated synthetic spectrum
  @param    vins     Instrumental broadening in km/s
  @param    vrot     Stellar rotational velocity in km/s
  @param    vmac     Stellar macroturbulent velocity in km/s
  @param    teff     Stellar effective temperature in K
  @return   CPL_ERROR_NONE iff OK

  @doc      
  This function convolves the synthetic spectrum with broadening profiles
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_convol_comp(cpl_array     *wobs,
			                     cpl_array     *fsyn,
			                     cpl_array     *fcon,
                                  double        vins,
                                  double        vrot,
                                  double        vmac,
                                  double        teff)

{
    cpl_errorstate prestate = cpl_errorstate_get();

/*
 * This function performs the convolution of the interpolated synthetic   *
 * spectrum with broadening profiles for macroturbulent velocity,         *
 * projected rotational velocity (vsini) and instrumental broadening      *
 * given for the three observing modes :                                  *
 * - singleUHR (ultra-high resolution - one UT),                          *
 * - singleHR (high-resolution - one UT) and                              *
 * - multiHR (high-resolution - various UT)                               *
*/

// Defining the number of elements of the arrays

    cpl_size nobs = cpl_array_get_size (wobs);

// Dispersion of wavelength array in km/s and central wavelength
    double vlight = 299792.458 ;

    double wini = cpl_array_get_double (wobs, 0, NULL);
    int    nfin = (int)nobs-1 ;
    double wfin = cpl_array_get_double (wobs, nfin, NULL);
    double wmid = sqrt (wini * wfin) ;

    double w1 = cpl_array_get_double (wobs, 5, NULL);
    double w2 = cpl_array_get_double (wobs, 6, NULL);
    double vdw = (w2 - w1) / w1 * vlight ;
    //double xdw = (wfin - wini) / (double)nobs ;
    double xdw = (wfin - wini) / (double)nfin ;

    //double w1b = cpl_array_get_double (wobs, 599, NULL);
    //double w2b = cpl_array_get_double (wobs, 600, NULL);
    //double vdwb = (w2b - w1b) / w1b * vlight ;

    //double wfinb = cpl_array_get_max(wobs);
    //printf("Wobs_ini fin: %f %f %f \n", wini, wfin, wfinb);

    //printf("Number of points of observed spectrum: %i \n", (int)nobs);
    //printf("Wavelength dispersion in nm : %f \n", xdw);
    //printf("Central Wavelength in nm : %f \n", wmid);
    //printf("Wavelength dispersion in km/s : %f \n", vdw);

    //printf("Wavelength dispersion in km/s : %f \n", vdwb);

// We define the arrays

    cpl_array *fsyc = cpl_array_new (nobs, CPL_TYPE_DOUBLE);
    cpl_array *xkn=NULL;
    cpl_array *ykn=NULL;

// We perform the convolution

    for (int i = 0; i < (int)nobs; i++) {
       double cfs =  cpl_array_get_double (fsyn, i, NULL);
       cpl_array_set_double (fcon, i, cfs);
       cpl_array_set_double (fsyc, i, cfs);
    }

    //We select the broadening profile

    //Macroturbulent velocity

    if (vmac > 2.*vdw) {
      double vbr = vmac ;
      int ibr = 1 ;

      cpl_ensure_code(espda_convol_prof(wobs, &xkn, &ykn, vbr, teff, wmid, ibr)
                            == CPL_ERROR_NONE, cpl_error_get_code());

      cpl_size npk = cpl_array_get_size (xkn);

      //Total area of the convolution profile
      double tfy = 0;
      for (int j = 0; j < (int)npk; j++) {
   	  double cfk =  cpl_array_get_double (ykn, j, NULL);
   	  tfy = tfy + cfk ;
      }

      //Normalization factor
   	  tfy = tfy * xdw;
   	  //printf("Normalization factor : %f \n", tfy);

      for (int i = 0; i < (int)nobs; i++) {

    	 double cfs = 1.;
    	 cpl_array_set_double (fcon, i, cfs);

    	 if (i >= (int)npk && i < (int)nobs-(int)npk) {

           double cfy = 0;
           int j = 0;
           for (int k = i-(int)npk/2; k < i+(int)npk/2; k++) {
        	  double cfs =  cpl_array_get_double (fsyc, k, NULL);
        	  double cfk =  cpl_array_get_double (ykn, j, NULL);
        	  cfy = cfy + (cfs * cfk) * xdw;
        	  j++;
           }
           cfy = cfy / tfy ;
           cpl_array_set_double (fcon, i, cfy);
         }
      }

      //We reinitiate the convolved synthetic spectrum to use it as input
      for (int i = 0; i < (int)nobs; i++) {
         double cfc =  cpl_array_get_double (fcon, i, NULL);
         cpl_array_set_double (fsyc, i, cfc);
      }
    }

    //Projected rotational velocity (vsini)

    if (vrot > 2.*vdw) {
      double vbr = vrot ;
      int ibr = 2 ;

      cpl_ensure_code(espda_convol_prof(wobs, &xkn, &ykn, vbr, teff, wmid, ibr)
                            == CPL_ERROR_NONE, cpl_error_get_code());

      cpl_size npk = cpl_array_get_size (xkn);

      //Total area of the convolution profile
      double tfy = 0;
      for (int j = 0; j < (int)npk; j++) {
   	  double cfk =  cpl_array_get_double (ykn, j, NULL);
   	  tfy = tfy + cfk ;
      }

      //Normalization factor
   	  tfy = tfy * xdw;
   	  //printf("Normalization factor : %f \n", tfy);

      for (int i = 0; i < (int)nobs; i++) {

    	 double cfs = 1.;
    	 cpl_array_set_double (fcon, i, cfs);

    	 if (i >= (int)npk && i < (int)nobs-(int)npk) {

           double cfy = 0;
           int j = 0;
           for (int k = i-(int)npk/2; k < i+(int)npk/2; k++) {
        	  double cfs =  cpl_array_get_double (fsyc, k, NULL);
        	  double cfk =  cpl_array_get_double (ykn, j, NULL);
        	  cfy = cfy + (cfs * cfk) * xdw;
        	  j++;
           }
           cfy = cfy / tfy ;
           cpl_array_set_double (fcon, i, cfy);
         }
      }

      //We reinitiate the convolved synthetic spectrum to use it as input
      for (int i = 0; i < (int)nobs; i++) {
         double cfc =  cpl_array_get_double (fcon, i, NULL);
         cpl_array_set_double (fsyc, i, cfc);
      }
    }

    //Instrumental broadening

    if (vins > 2.*vdw) {
      double vbr = vins ;
      int ibr = 3 ;

      cpl_ensure_code(espda_convol_prof(wobs, &xkn, &ykn, vbr, teff, wmid, ibr)
                            == CPL_ERROR_NONE, cpl_error_get_code());

      cpl_size npk = cpl_array_get_size (xkn);

      //Total area of the convolution profile
      double tfy = 0;
      for (int j = 0; j < (int)npk; j++) {
   	  double cfk =  cpl_array_get_double (ykn, j, NULL);
   	  tfy = tfy + cfk ;
      }

      //Normalization factor
   	  tfy = tfy * xdw;
   	  //printf("Normalization factor : %f \n", tfy);

      for (int i = 0; i < (int)nobs; i++) {

    	 double cfs = 1.;
    	 cpl_array_set_double (fcon, i, cfs);

    	 if (i >= (int)npk && i < (int)nobs-(int)npk) {

           double cfy = 0;
           int j = 0;
           for (int k = i-(int)npk/2; k < i+(int)npk/2; k++) {
        	  double cfs =  cpl_array_get_double (fsyc, k, NULL);
        	  double cfk =  cpl_array_get_double (ykn, j, NULL);
        	  cfy = cfy + (cfs * cfk) * xdw;
        	  j++;
           }
           cfy = cfy / tfy ;
           cpl_array_set_double (fcon, i, cfy);
         }
      }
    }

    // Closing the function

    if (!cpl_errorstate_is_equal(prestate)) {
      return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
               "Unable to perform the convolution...");
    }

    //Delete arrays before closing
    cpl_array_delete(xkn);
    cpl_array_delete(ykn);
    cpl_array_delete(fsyc);

    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Create broadening profiles
  @param    xobs   Array of wavelength of the observed spectrum
  @param    xk     Array of x-values of the broadening profile
  @param    yk     Array of y-values of the broadening profile
  @param    vbr    Broadening velocity of the profile in km/s
  @param    teff   Stellar effective temperature in K
  @param    wmid   Central wavelength in the logaritmic scale in nm
  @param    ibr    Index to select the type of broadening profile
  @return   CPL_ERROR_NONE iff OK

  @doc
  This function creates the selected broadening profile
*/

/*----------------------------------------------------------------------------*/

cpl_error_code espda_convol_prof(cpl_array     *xobs,
		                         cpl_array      **xk,
		                         cpl_array      **yk,
							     double          vbr,
							     double         teff,
								 double         wmid,
		                         int             ibr)

{
    cpl_errorstate prestate = cpl_errorstate_get();

    // Defining the number of elements of the arrays
    cpl_size   npoint = cpl_array_get_size (xobs);
    int        npt = (int)npoint ;

    //Macroturbulence profile: Gaussian (Radial-tangential) profile :  ibr = 1
    //Rotation profile: Gray's profile:  ibr = 2
    //Instrumental profile: Gaussian Kernel :  ibr = 3

    //Macroturbulent velocity

    if ((int)ibr == 1) {
      printf("Macroturbulence= %f %i \n",vbr, ibr);
      double vlight = 299792.458 ;
      //double resl = vlight/vbr/CPL_MATH_FWHM_SIG ;
      double resl = vlight/vbr ;
      double fwhm = wmid/resl ;
	  //CPL_MATH_FWHM_SIG: FWHM per Sigma, 2.0*sqrt(2.0*log(2.0))
      double sigma = fwhm/CPL_MATH_FWHM_SIG ;
	  double a1 = 0. ;
	  double a2 = sigma ;
      //double sigma = fwhm/2./SQRT(2.*alog(2.))  ;
      //double a0= 1./SQRT(2.*CPL_MATH_PI)/sigma  ;
	  //CPL_MATH_SQRT2PI: sqrt(2pi)
	  double a0= 1./CPL_MATH_SQRT2PI/sigma ;

	  double xu = 5.*sigma ;
	  double xd = -5.*sigma ;

      //printf("Macro G prof= %i %f %f %f %f \n",ibr, a0, a1, a2, sigma);
    // We create the xk array of the gaussian kernel
	  int nk= 0;
	  for (int i = 0; i < (int)npt; i++){
		  double cxo = cpl_array_get_double (xobs, i, NULL);
		  double cxn = cxo - wmid ;
		  if (cxn > xd && cxn < xu) {
           nk++ ;
		  }
	  }

	  int npk = nk ;
	  *xk = cpl_array_new(npk, CPL_TYPE_DOUBLE);
	  *yk = cpl_array_new(npk, CPL_TYPE_DOUBLE);
	  cpl_array *xkn = cpl_array_new(npk, CPL_TYPE_DOUBLE);
	  cpl_array *ykn = cpl_array_new(npk, CPL_TYPE_DOUBLE);

	  int k = 0 ;
	  for (int i = 0; i < (int)npt; i++){
		  double cxo = cpl_array_get_double (xobs, i, NULL);
		  double cxn = cxo - wmid ;
		  if (cxn > xd && cxn < xu) {
		   cpl_array_set_double (xkn, k, cxn);
		   //printf("Macro prof= %i %i %f %f %f \n", i, k, xd, xu, cxn);
		   k++;
		  }
	  }

	// We create the yk array of the gaussian kernel
	  for (int k = 0; k < (int)npk; k++){
		 double xi = cpl_array_get_double (xkn, k, NULL) ;
         double yi = a0 * exp (-(((xi-a1)*(xi-a1))/(a2*a2))/2.) ;
	     cpl_array_set_double (ykn, k, yi);
         //printf("Macro prof= %i %f %f \n", k, xi, yi);
     }

	  for (int j = 0 ; j < (int)npk; j++) {
	    double cx =  cpl_array_get_double (xkn, j, NULL);
	    double cy =  cpl_array_get_double (ykn, j, NULL);
	    cpl_array_set_double (*xk, j, cx);
	    cpl_array_set_double (*yk, j, cy);
	  }
      //Deleting arrays and close the function
      cpl_array_delete(xkn);
      cpl_array_delete(ykn);
    }

    //Projected rotational velocity

    if ((int)ibr == 2) {
      printf("Rotation= %f %i \n",vbr,ibr);
      double vlight = 299792.458 ;
      double resl = vlight/vbr/CPL_MATH_FWHM_SIG ;
      double fwhm = wmid/resl ;

	  //CPL_MATH_FWHM_SIG: FWHM per Sigma, 2.0*sqrt(2.0*log(2.0))
      double sigma = fwhm/CPL_MATH_FWHM_SIG ;
	  //double a1 = 0. ;
	  //double a2 = sigma ;
      //double sigma = fwhm/2./SQRT(2.*alog(2.))  ;
      //double a0= 1./SQRT(2.*CPL_MATH_PI)/sigma  ;
	  //CPL_MATH_SQRT2PI: sqrt(2pi)
	  //double a0= 1./CPL_MATH_SQRT2PI/sigma ;

	  double xu = 5.*sigma ;
	  double xd = -5.*sigma ;

      //Defining the Gray profile
	  //CPL_MATH_PI   3.14159265358979
	  double xg = vbr/vlight*wmid ;

	  //Approximation of Linear limb darkening law at wavelength ~ 600nm
	  double ldc = 0.933083 - teff * 0.0000592331 ;

      double d  = 1.0/(CPL_MATH_PI*(1.0-ldc/3.0)) ;
      double c1 = 2.0*(1.0-ldc)*d ;
      double c2 = 0.5*CPL_MATH_PI*ldc*d ;

    // We create the xk array of the gaussian kernel
	  int nk= 0;
	  for (int i = 0; i < (int)npt; i++){
		  double cxo = cpl_array_get_double (xobs, i, NULL);
		  double cxn = cxo - wmid ;
		  if (cxn > xd && cxn < xu) {
           nk++ ;
		  }
	  }

	  int npk = nk ;
	  *xk = cpl_array_new(npk, CPL_TYPE_DOUBLE);
	  *yk = cpl_array_new(npk, CPL_TYPE_DOUBLE);
	  cpl_array *xkn = cpl_array_new(npk, CPL_TYPE_DOUBLE);
	  cpl_array *ykn = cpl_array_new(npk, CPL_TYPE_DOUBLE);

	  int k = 0 ;
	  for (int i = 0; i < (int)npt; i++){
          double cxo = cpl_array_get_double (xobs, i, NULL);
		  double cxn = cxo - wmid ;
		  if (cxn > xd && cxn < xu) {
		   cpl_array_set_double (xkn, k, cxn);
		   //printf("Macro prof= %i %i %f %f %f \n", i, k, xd, xu, cxn);
		   k++;
		  }
	  }

	// We create the yk array of the Gray's profile kernel

	  //for (int k = 0; k < (int)npk; k++){
	  //  double xi = cpl_array_get_double (xkn, k, NULL) ;
      //  double xu  = 1.0-((xi/xg)*(xi/xg))
	  //  double yi = c1 * sqrt(xu) + c2 * xu
	  //  cpl_array_set_double (ykn, k, yi);
      //  //printf("Macro prof= %i %f %f \n", k, xi, yi);
      //}

	  cpl_array *xn = cpl_array_duplicate (xkn);
	  for (int k = 0; k < (int)npk; k++){
	    double xi = cpl_array_get_double (xkn, k, NULL) ;
        double cn  = 1.0-((xi/xg)*(xi/xg)) ;
        cpl_array_set_double (xn, k, cn);
	  }

	  cpl_array *xa = cpl_array_duplicate (xn);
	  double cxa = 0.5 ;
	  cpl_array_power (xa, cxa);
	  cpl_array_multiply_scalar (xa, c1);
	  cpl_array *xb = cpl_array_duplicate (xn);
	  cpl_array_multiply_scalar (xb, c2);
	  cpl_array *xc = cpl_array_duplicate (xa);
	  cpl_array_add (xc, xb);

	  for (int k = 0; k < (int)npk; k++){
	    double yn = cpl_array_get_double (xc, k, NULL) ;
	    cpl_array_set_double (ykn, k, yn);
	  }

	  for (int j = 0 ; j < (int)npk; j++) {
	    double cx =  cpl_array_get_double (xkn, j, NULL);
	    double cy =  cpl_array_get_double (ykn, j, NULL);
	    cpl_array_set_double (*xk, j, cx);
	    cpl_array_set_double (*yk, j, cy);
	  }
      //Deleting arrays and close the function
	  cpl_array_delete(xn);
      cpl_array_delete(xa);
      cpl_array_delete(xb);
      cpl_array_delete(xc);
      cpl_array_delete(xkn);
      cpl_array_delete(ykn);
    }

    //Instrumental broadening

    if ((int)ibr == 3) {
      printf("Resolution= %f %i \n",vbr,ibr);
      double vlight = 299792.458 ;
      double resl = vlight/vbr ;
      double fwhm = wmid/resl ;
	  //CPL_MATH_FWHM_SIG: FWHM per Sigma, 2.0*sqrt(2.0*log(2.0))
      double sigma = fwhm/CPL_MATH_FWHM_SIG ;
	  double a1 = 0. ;
	  double a2 = sigma ;
      //double sigma = fwhm/2./SQRT(2.*alog(2.))  ;
      //double a0= 1./SQRT(2.*CPL_MATH_PI)/sigma  ;
	  //CPL_MATH_SQRT2PI: sqrt(2pi)
	  double a0= 1./CPL_MATH_SQRT2PI/sigma ;

	  double xu = 5.*sigma ;
	  double xd = -5.*sigma ;

    // We create the xk array of the gaussian kernel
	  int nk= 0;
	  for (int i = 0; i < (int)npt; i++){
		  double cxo = cpl_array_get_double (xobs, i, NULL);
		  double cxn = cxo - wmid ;
		  if (cxn > xd && cxn < xu) {
           nk++ ;
		  }
	  }

	  int npk = nk ;
	  *xk = cpl_array_new(npk, CPL_TYPE_DOUBLE);
	  *yk = cpl_array_new(npk, CPL_TYPE_DOUBLE);
	  cpl_array *xkn = cpl_array_new(npk, CPL_TYPE_DOUBLE);
	  cpl_array *ykn = cpl_array_new(npk, CPL_TYPE_DOUBLE);

	  int k = 0 ;
	  for (int i = 0; i < (int)npt; i++){
		  double cxo = cpl_array_get_double (xobs, i, NULL);
		  double cxn = cxo - wmid ;
		  if (cxn > xd && cxn < xu) {
		   cpl_array_set_double (xkn, k, cxn);
		   //printf("Macro prof= %i %i %f %f %f \n", i, k, xd, xu, cxn);
		   k++;
		  }
	  }

	// We create the yk array of the gaussian kernel
	  for (int k = 0; k < (int)npk; k++){
		 double xi = cpl_array_get_double (xkn, k, NULL) ;
         double yi = a0 * exp (-(((xi-a1)*(xi-a1))/(a2*a2))/2.) ;
	     cpl_array_set_double (ykn, k, yi);
         //printf("Macro prof= %i %f %f \n", k, xi, yi);
     }
	  for (int j = 0 ; j < (int)npk; j++) {
	    double cx =  cpl_array_get_double (xkn, j, NULL);
	    double cy =  cpl_array_get_double (ykn, j, NULL);
	    cpl_array_set_double (*xk, j, cx);
	    cpl_array_set_double (*yk, j, cy);
	  }
      //Deleting arrays and close the function
      cpl_array_delete(xkn);
      cpl_array_delete(ykn);
     }

    // Closing the function

        if (!cpl_errorstate_is_equal(prestate)) {
          return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
                   "Unable to create the convolution profile...");
        }

        return CPL_ERROR_NONE;
    }

/**@}*/
