/* 
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Id: espda_fit.c 161538 2014-05-22 13:56:10Z gcupani $
 * $Author: gcupani $
 * $Date: 2014-05-22 15:56:10 +0200 (Thu, 22 May 2014) $
 * $Revision: 161538 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include "espda_fit.h"
#include <gsl/gsl_version.h>

/*----------------------------------------------------------------------------*/
/**
  @defgroup   espda_fit  Curve fitting
  
  Spectra can be fitted with cubic splines (to define the level of continuum 
  emission) or with Voigt profiles (to interpret the absorption lines). In both
  cases, the fitting curve is a function of wavelength.
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Fit a cubic spline to a spectrum.
  @param    ref   A reference spectrum (SPEC_FIT).
  @param    col   The column with the fit.
  @param    spec  The spectrum to which the fit is applied (SPEC_FIT).
  @param    line  The list of lines to which the fit is applied (LINE_VOIGT;
                  may be NULL).
  @return   CPL error code.  

  This function takes a reference spectrum @a ref, fits a cubic spline on its 
  flux, and applies the fitting curve to another spectrum @a spec and 
  (optionally) to a list of lines @line. @a ref is typically an undersampled 
  version of @a spec. The fit is saved as a separate column @a col in @a ref, 
  @a spec, and (optionally) @a line. Wavelengths not strictly contained in the
  range of @a ref are flagged as invalid and not fitted.

  The cubic spline is computed by the GSL package @p gsl_spline. If @a ref has 
  only two rows, a two-point interpolation on flux values is performed; if it 
  has only one row, the flux value is taken as fit value.
  
  If @a ref or @a spec are NULL, an error is issued. If @a col does not exist, 
  an error is issued. If @a line is NULL and verbosity level is greater than 0, 
  a warning is issued. 
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_fit_cspline(cpl_table *ref, char *col, cpl_table *spec,
    cpl_table *line) {
    extern int v_level;
    extern char *grp_mode;    
    espda_util_error(ref  == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    espda_util_error(spec == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    espda_util_error(!cpl_table_has_column(spec, col), CPL_ERROR_NULL_INPUT, 
        cpl_error_get_code()); 
    espda_util_warning(line == NULL, "list of lines not provided.");
    espda_struct_init(spec, col, CPL_TYPE_DOUBLE);
    espda_util_safe(line, espda_struct_init(line, col, CPL_TYPE_DOUBLE));                  
    cpl_size ref_size  = cpl_table_get_nrow(ref);
    double wavel_min = cpl_table_get_double(ref, "WAVEL", 0, NULL);
    double wavel_max = cpl_table_get_double(ref, "WAVEL", ref_size - 1, NULL);
    cpl_size line_size = 0;
    cpl_table *spec_rej = NULL;

    // Usable regions of the spectrum are selected
    if ((int)ref_size > 1) {
        spec_rej = espda_struct_extract(spec, 2, "WAVEL", CPL_NOT_GREATER_THAN,
            wavel_min, "WAVEL", CPL_NOT_LESS_THAN, wavel_max);
        espda_struct_erase(spec, 2, "WAVEL", CPL_NOT_GREATER_THAN, wavel_min,
            "WAVEL", CPL_NOT_LESS_THAN, wavel_max);
        if (line != NULL) {
	    espda_struct_erase(line, 2, "WAVEL", CPL_NOT_GREATER_THAN, 
	        wavel_min, "WAVEL", CPL_NOT_LESS_THAN, wavel_max);
            line_size = cpl_table_get_nrow(line);
        }
    }
    espda_util_safe(line, line_size = cpl_table_get_nrow(line));
    cpl_size spec_size = cpl_table_get_nrow(spec);

    // Single-row spectrum
    if ((int)ref_size == 1) {
        espda_struct_copy(ref, col, 0, ref, "FLUX", 0);
        cpl_table_fill_column_window_double(spec, col, 0, spec_size,
            cpl_table_get_double(ref, "FLUX", 0, NULL));
        espda_util_safe(line, cpl_table_fill_column_window_double(line, 
            col, 0, line_size, cpl_table_get_double(ref, "FLUX", 0, NULL)));
    }

    // Multiple-row spectrum
    else {
        double pts_x[(int)ref_size], pts_y[(int)ref_size]; 
        for (int i = 0; i < (int)ref_size; i++) {
            pts_x[i] = cpl_table_get_double(ref, "WAVEL", i, NULL);
            pts_y[i] = cpl_table_get_double(ref, "FLUX",  i, NULL);
            cpl_table_set_double(ref, col,  i, pts_y[i]);
        }
        
        // Two-point interpolation
        if ((int)ref_size == 2) {
            for (int j = 0; j < spec_size; j++) {
                cpl_table_set_double(spec, col, j, pts_y[0] +
                    (pts_y[1] - pts_y[0]) / (pts_x[1] - pts_x[0]) *
                    (cpl_table_get_double(spec, "WAVEL", j, NULL) - pts_x[0]));
            }
            if (line != NULL) {
                for (int j = 0; j < line_size; j++) {
                    cpl_table_set_double(line, col, j, pts_y[0] +
                        (pts_y[1] - pts_y[0]) / (pts_x[1] - pts_x[0]) *
                        (cpl_table_get_double(line, "WAVEL", j, NULL) 
                        - pts_x[0]));
                }
            }
        }

        // Spline interpolation
        else if ((int)ref_size > 2) {
            gsl_interp_accel *acc = gsl_interp_accel_alloc();
            gsl_spline *spline = gsl_spline_alloc(gsl_interp_cspline, 
                (int)ref_size);
            gsl_spline_init(spline, pts_x, pts_y, (int)ref_size); 
            for (int j = 0; j < spec_size; j++) {
                cpl_table_set_double(spec, col, j, gsl_spline_eval(spline, 
                    cpl_table_get_double(spec, "WAVEL", j, NULL), acc));
            }
            if (line != NULL) {
                for (int j = 0; j < line_size; j++) {
                    cpl_table_set_double(line, col, j, gsl_spline_eval(spline, 
                        cpl_table_get_double(line, "WAVEL", j, NULL), acc));
                }
            }
            gsl_interp_accel_free(acc);
            gsl_spline_free(spline);
        }
    }
    
    // Wavelengths not covered by the reference spectrum are flagged as invalid     
    if (strcmp(grp_mode, "OVER")) {
        for (int i = 0; i < spec_size; i++) {
            double wavel_min = cpl_table_get_double(spec, "WAVEL", i, NULL) -
                cpl_table_get_double(spec, "PIXSIZE", i, NULL) * 0.5;
            double wavel_max = cpl_table_get_double(spec, "WAVEL", i, NULL) +
                cpl_table_get_double(spec, "PIXSIZE", i, NULL) * 0.5;
            cpl_table *temp = espda_struct_extract(ref,  1, "WAVEL", 
                CPL_NOT_LESS_THAN,    wavel_min);
            cpl_table *extr = espda_struct_extract(temp, 1, "WAVEL", 
                CPL_NOT_GREATER_THAN, wavel_max);
            if (cpl_table_get_nrow(extr) <= 0) {
                cpl_table_set_invalid(spec, col, i);
            }
            espda_struct_delete(CPL_TABLE, 2, temp, extr);
        }
    }
    cpl_table_set_column_invalid(spec_rej, col, 0, 
        cpl_table_get_nrow(spec_rej));
    cpl_table_insert(spec, spec_rej, spec_size);
    cpl_table_delete(spec_rej);
    espda_struct_sort(spec, 1, "WAVEL", FALSE);    
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
 */
/*----------------------------------------------------------------------------*/

//cpl_table *espda_spec_rebin(cpl_table *fspec, int kappa, double wavel_start,
//    double wavel_end, double vel_step) {
/*
cpl_table *espda_fit_sg(cpl_table *spec_in, char *col, int width) {

    cpl_table *spec_out = cpl_table_duplicate(spec_in);
    int spec_size = cpl_table_get_nrow(spec_in);
    //double spec_wavel[spec_size];
    double spec_data[spec_size];
    for (int i = 0; i < spec_size; i++) {
      //spec_wavel[i] = cpl_table_get_double(spec_in, "WAVEL", i, NULL);
        spec_data[i] = cpl_table_get_double(spec_in, "FLUX", i, NULL); 
    }  
    double spec_filt[spec_size];
    //char out = sgfilter(spec_data, spec_filt, spec_size, width, width, 0, 10);
    for (int i = 0; i < spec_size; i++) {
        cpl_table_set_double(spec_out, col, i, spec_filt[i]);
    }
    
    return spec_out;
}
*/

/*----------------------------------------------------------------------------*/
/**
 */
/*----------------------------------------------------------------------------*/

//cpl_table *espda_spec_rebin(cpl_table *fspec, int kappa, double wavel_start,
//    double wavel_end, double vel_step) {

/*
void *espda_fit_callback(const size_t iter, void *params, 
			const gsl_multifit_nlinear_workspace *w) {
     printf("%i\n", iter);
}
*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Fit Voigt profiles to absorption lines.
  @param    spec_over   A spectrum over-chunk (SPEC_FIT).
  @param    spec_chunk  A spectrum chunk (SPEC_FIT).
  @param    line_chunk  The list of lines in the spectrum chunk (LINE_VOIGT).
  @param    prof_chunk  The instrument profile in the spectrum chunk.
  @param    ion         A list of ionic transitions (ION).
  @param    voigt_func  The tabulated Voigt function (VOIGT).
  @param    rchisq      The reduced chi-square from fitting.
  @param    iter_arr    The number of fitting iterations.
  @param    iter_max    The maximum number of iterations.
  @param    spec_out    The fitted spectrum chunk (SPEC_FIT; must be NULL).
  @param    line_out    The list of fitted lines (LINE_VOIGT; must be NULL).
  @return   CPL error code.  
  
  This function takes a spectrum chunk @a spec_chunk and a list of lines 
  @a line_chunk, fits the lines with Voigt profiles, and produces a fitted 
  spectrum @a spec_out and a list of lines with fitted Voigt parameters 
  @a line_out. The Voigt fitting is based on a previous identification of the 
  lines that are to be fitted and takes the wavelength, the oscillator strength, 
  and the damping parameter of the identified lines from a list of ionic 
  transitions @a ion. The spectral chunk is convolved with the instrument 
  profile @a prof_chunk (a slightly-larger over-chunk @a spec_over is used 
  instead of @a spec_chunk to avoid border effects). An optional table 
  @a voigt_func with previously-computed values of the Voigt function can be 
  used to speed up the computation. The fitting procedure is terminated when 
  either the algorithm converges or a maximum number of iterations is reached 
  @a iter_max. 
  
  The function behaves differently depending on external variables. Four modes
  are available:
  
  -# standard fit on a rebinned spectrum (@a grp_max != @p "MAX", @a cycle_flag
     = @p "rebinned"): fitting is performed and results are saved in @a spec_out
     and @a line_out;
  -# skipped fit on a non-rebinned spectrum (@a grp_max != @p "MAX", 
     @a cycle_flag != @p "rebinned"): fitting is skipped and results from a 
     previous fitting of the rebinned spectrum are saved in @a spec_out and 
     @a line_out;
  -# last run on a rebinned spectrum (@a grp_max = @p "MAX", @a cycle_flag
     = @p "rebinned"): fitting is skipped; the rebinned @a spec_out is copied 
     from @a spec_chunk and and a reduced chi-square is computed by applying on 
     @a spec_chunk the profile obtained the Voigt parameters in @a line_chunk;
  -# last run on a non-rebinned spectrum (@a grp_max = @p "MAX", @a cycle_flag
     != @p "rebinned"): same as 3, but on a non-rebinnd @a spec_chunk instead.
  
  The values of reduced chi-square from fitting and the number of fitting
  iterations are saved in the 4-element arrays @a rchisq and @a iter_arr, 
  respectively, depending on the chosen mode (values from mode 1 are saved in 
  the first element, etc.).
  
  The fitting is performed by the GSL package @p gsl_multifit_nlin using the 
  Levenberg-Marquardt algorithm. For details on the convolution, see 
  @p espda_spec_conv.
  
  If the @a spec_over, @a spec_chunk, @a line_chunk, @a prof_chunk, or @a ion 
  are NULL, an error is issued. If @a voigt_func is NULL and verbosity level is 
  greater than 0, a warning is issued. 
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_fit_voigt(cpl_table *spec_over, cpl_table *spec_chunk,
    cpl_table *line_chunk, cpl_table *prof_chunk, cpl_table *ion, 
    cpl_table *voigt_func, cpl_array *rchisq, cpl_array *iter_arr, 
    cpl_size iter_max, cpl_table **spec_out, cpl_table **line_out) {
    extern int v_level;
    extern int cycle;
    extern char *grp_flag, *cycle_flag;
    
    espda_util_error(spec_over == NULL, CPL_ERROR_NULL_INPUT, 
		     cpl_error_get_code());
    espda_util_error(spec_chunk == NULL, CPL_ERROR_NULL_INPUT, 
		     cpl_error_get_code());
    espda_util_error(line_chunk == NULL, CPL_ERROR_NULL_INPUT, 
		     cpl_error_get_code());
    espda_util_error(prof_chunk == NULL, CPL_ERROR_NULL_INPUT, 
		     cpl_error_get_code());
    espda_util_error(ion  == NULL, CPL_ERROR_NULL_INPUT, 
		     cpl_error_get_code());
    espda_util_warning(voigt_func == NULL, "tabulated Voigt function not "
		       "provided.");

    // Information about ionic transitions is extracted
    cpl_array *ion_wavel  = NULL;
    cpl_array *ion_oscs   = NULL; 
    cpl_array *ion_damp   = NULL; 
    espda_line_ion(line_chunk, ion, &ion_wavel, &ion_oscs, &ion_damp); 

    // Structures with Voigt parameters are created
    cpl_array  *par_pos  = NULL, *par_fix = NULL;
    cpl_table  *par_lim  = NULL;
    gsl_vector *par_vary = NULL;
    cpl_size   line_size = cpl_table_get_nrow(line_chunk);
    cpl_size    vary_num, par_num = 4 * line_size;
    espda_voigt_create(line_chunk, &par_pos, &par_fix, &par_vary, &par_lim, 
        &vary_num);
    //gsl_vector_fprintf(stdout, par_vary, "%f");
    
    // GSL multifit is initialized
    size_t   spec_size = (size_t)cpl_table_get_nrow(spec_chunk);

    //######
    struct gsl_multifit_data spec_data = {spec_over, spec_chunk, line_chunk,
        prof_chunk, voigt_func, par_fix, par_pos, par_lim, par_num, ion_wavel, 
        ion_oscs, ion_damp};
    /*
    double *spec_fit = cpl_table_get_data_double(spec_chunk, "FIT");
    cpl_array *spec_norm = cpl_array_new(spec_size, CPL_TYPE_DOUBLE);
    cpl_array *spec_normerr = cpl_array_new(spec_size, CPL_TYPE_DOUBLE);
    cpl_array *spec_cont = cpl_array_new(spec_size, CPL_TYPE_DOUBLE);
    cpl_array_copy_data_double(spec_norm,
			       cpl_table_get_data_double(spec_chunk, "FLUX"));
    cpl_array_copy_data_double(spec_normerr,
			       cpl_table_get_data_double(spec_chunk, "FLUXERR"));
    cpl_array_copy_data_double(spec_cont,
			       cpl_table_get_data_double(spec_chunk, "CONT"));
    cpl_array_divide(spec_norm, spec_cont);
    cpl_array_divide(spec_normerr, spec_cont);
    struct gsl_multifit_data spec_data_new = {spec_over, spec_fit, spec_norm,
					    spec_normerr, line_chunk,
        prof_chunk, voigt_func, par_fix, par_pos, par_lim, par_num, ion_wavel, 
        ion_oscs, ion_damp};
    */
    //######
    /*
    struct gsl_multifit_data spec_data = {spec_over, spec_chunk, line_chunk,
        prof_chunk, voigt_func, par_fix, par_pos, par_lim, par_num, ion_wavel, 
        ion_oscs, ion_damp};
    */
    int dof = spec_size - vary_num;
    if (iter_max == 0) {
        dof = spec_size; 
    }

    // This is just a patch! Sometimes the range is too small; in this way
    // it is fitted anyway (but quality is not assured)
    if ((int)spec_size < (int)vary_num) {
        spec_size = vary_num;
    }
    const gsl_multifit_fdfsolver_type *type = gsl_multifit_fdfsolver_lmsder;

    gsl_multifit_fdfsolver *solver = gsl_multifit_fdfsolver_alloc(type, 
					  spec_size, vary_num);
    gsl_multifit_function_fdf func;
    func.f   = &espda_model_transm_old;
    func.df  = NULL; 
    func.fdf = NULL;
    func.n = spec_size;
    func.p = vary_num;
    func.params = &spec_data;


    // This takes long!!
    //gsl_multifit_fdfsolver_set(solver, &func, par_vary);
    

    gsl_matrix *covar = NULL;
#if defined GSL_MAJOR_VERSION && GSL_MAJOR_VERSION >= 2
    gsl_matrix *J;
#endif

    /*****/

    /*
    gsl_multifit_nlinear_fdf N_func;
    gsl_multifit_nlinear_parameters params = 
        gsl_multifit_nlinear_default_parameters();

    N_func.f = &espda_model_costf;
    N_func.df = NULL; //&espda_model_costdf; 
    N_func.fvv = NULL;
    N_func.n = spec_size;
    N_func.p = vary_num;
    N_func.params = &spec_data;

    const gsl_multifit_nlinear_type *N_type = gsl_multifit_nlinear_trust;

    gsl_multifit_nlinear_workspace *work = 
        gsl_multifit_nlinear_alloc(N_type, &params, spec_size, vary_num);

    gsl_vector *f = gsl_multifit_nlinear_residual(work);
    gsl_vector *x = gsl_multifit_nlinear_position(work);
    int info;
    double chisq0, chisq, rcond, xsq;
    // This takes long!
    gsl_multifit_nlinear_init(par_vary, &N_func, work);

    gsl_matrix *jac = gsl_multifit_nlinear_jac(work);
    //printf("%i %i\n", jac->size1, jac->size2);
    //gsl_matrix_free(jac);

    gsl_matrix *jacin;
    gsl_vector *fit_conv;
    espda_model_transm(par_vary, &spec_data, &fit_conv, &jacin);
    //printf("%i %i\n", jacin->size1, jacin->size2);

    //for (int i = 0; i < 1000; i++) {
    //  printf("%e %e | %e %e | %e %e | %e %e\n", gsl_matrix_get(jac, i, 0), gsl_matrix_get(jacin, i, 0), gsl_matrix_get(jac, i, 1), gsl_matrix_get(jacin, i, 1), gsl_matrix_get(jac, i, 2), gsl_matrix_get(jacin, i, 2), gsl_matrix_get(jac, i, 3), gsl_matrix_get(jacin, i, 3));
    //}
    //double min, max, minin, maxin;
    //gsl_matrix_minmax(jac, &min, &max); 
    //gsl_matrix_minmax(jacin, &minin, &maxin);
    //printf("%e %e %e %e\n", min, max, minin, maxin);
    //gsl_matrix_free(jacin);
    //gsl_vector_free(fit_conv);
*/
/*
    gsl_multilarge_nlinear_fdf N_func;
    gsl_multilarge_nlinear_parameters params = 
        gsl_multilarge_nlinear_default_parameters();

    N_func.f = &espda_model_cost;
    N_func.df  = NULL; 
    N_func.fvv = NULL;
    N_func.n = spec_size;
    N_func.p = vary_num;
    N_func.params = &spec_data;

    const gsl_multilarge_nlinear_type *N_type = gsl_multilarge_nlinear_trust;

    gsl_multilarge_nlinear_workspace *work = 
        gsl_multilarge_nlinear_alloc(N_type, &params, spec_size, vary_num);

    gsl_vector *f = gsl_multilarge_nlinear_residual(work);
    gsl_vector *x = gsl_multilarge_nlinear_position(work);
    int info;
    double chisq0, chisq, rcond, xsq;
    gsl_multilarge_nlinear_init(par_vary, &N_func, work);
*/
    /*****/

    // When all groups are selected, the Voigt profiles are computed, not fitted
    if (!strcmp(grp_flag, "MAX")) {
        // This takes a lot!!
        gsl_multifit_fdfsolver_set(solver, &func, par_vary);

	*spec_out = cpl_table_duplicate(spec_chunk);
        *line_out = cpl_table_duplicate(line_chunk);
        cpl_array_set_double(rchisq, cycle + 2, RCHISQ);
	//cpl_array_set_double(rchisq, cycle + 2, 0.0);
    }

    // When a single group is selected, the Voigt profiles are fitted
    else {
        gsl_multifit_fdfsolver_set(solver, &func, par_vary);
	int iter = 0;
	if (!strcmp(cycle_flag, "rebinned ")) {
            // Minimization is performed
	  //cpl_table_dump(spec_chunk, 0, 3, NULL);
	    int iter_tot = 0;
	    int status;
            do {
	        if (iter > 0) {
#if defined GSL_MAJOR_VERSION && GSL_MAJOR_VERSION >= 2
                    int   info=0;
/*
                    status = gsl_multifit_nlinear_driver(iter_max, 
				 1e-2, 1e-1, 1e-2, NULL, NULL, &info, work);
*/
	  //printf("here\n");
/*
                    status = gsl_multilarge_nlinear_driver(iter_max, 
				 1e-2, 1e-1, 1e-2, NULL, NULL, &info, work);
*/

/*
		    status = gsl_multifit_fdfsolver_driver(solver, iter_max, 
							   1e-2, 1e-1, 1e-2, &info);
*/


		    clock_t start = clock();
		    double limit = 1.0;
		    for (int i = 0; i < iter_max; i++) {
		        gsl_multifit_fdfsolver_iterate(solver);
			gsl_multifit_fdfsolver_test(solver, 1e-2, 1e-1, 1e-2, 
						    &info);
			clock_t diff = (clock() - start) \
			               * 1000 / CLOCKS_PER_SEC;
			if (diff > limit * 1e3) {
			    cpl_msg_warning(cpl_func, "Line fitting exceeded "
					    "%3.2f seconds: exiting...", limit);
			    break;
			}
		    }


#else
                    status = gsl_multifit_fdfsolver_driver(solver, iter_max, 
							   1e-2, 1e-1);
#endif                    
                }
                else {
                    status = GSL_CONTINUE;
                }
                iter++;
                iter_tot++;
            } while (status == GSL_CONTINUE && iter < 2);
	    if (vary_num > 0) {
               	covar = gsl_matrix_alloc(vary_num, vary_num);


#if defined GSL_MAJOR_VERSION && GSL_MAJOR_VERSION >= 2
                J = gsl_matrix_alloc(spec_size, vary_num);
                gsl_multifit_fdfsolver_jac(solver, J);
                gsl_multifit_covar (J, 0.0, covar);
                gsl_matrix_free (J);

/*
                J = gsl_multifit_nlinear_jac(work);
                gsl_multifit_covar (J, 0.0, covar);
                gsl_matrix_free (J);
*/

//		gsl_multilarge_nlinear_rcond(&rcond, work);
#else 
               	gsl_multifit_covar(solver->J, 0.0, covar);
#endif

                espda_print_verb(1, cpl_msg_info(cpl_func, " Converged after "
                    "%i iterations!", iter));
               	if (vary_num < 12) {
                    espda_print_verb(3, espda_print_par(solver->x, par_vary, 
                        par_lim));
                    //espda_print_verb(3, espda_print_par(work->x, par_vary, 
                    //    par_lim));
                    espda_print_verb(3, espda_print_covar(covar, par_vary));
                }
            }
        }
        cpl_array *par_all = NULL;

        // Structures with Voigt parameters are merged
        espda_voigt_merge(line_chunk, ion_wavel, par_pos, par_fix, solver->x, 
            covar, par_lim, &par_all, NULL);
        //espda_voigt_merge(line_chunk, ion_wavel, par_pos, par_fix, work->x, 
        //    covar, par_lim, &par_all, NULL);
    	if (!strcmp(cycle_flag, "rebinned ")) {
            espda_struct_insert2(!strcmp(grp_flag, "MIN"), *line_out, 
                line_chunk);
        }
        cpl_array_set_double(rchisq,   cycle, RCHISQ); 
        cpl_array_set_int   (iter_arr, cycle, iter);
	//*spec_out = cpl_table_duplicate(spec_chunk);
        //*line_out = cpl_table_duplicate(line_chunk);
    }

    //######
    //cpl_table_copy_data_double(spec_chunk, "FIT", spec_fit);
    //######
    char *part_name = "chunk:   ";
    espda_print_verb(1, cpl_msg_info(cpl_func, " Reduced chi-squared on %s %s "
        "%7.4f (size: %4i, degrees of freedom: %4i)", cycle_flag, part_name, 
        RCHISQ, (int)spec_size, dof));

    espda_struct_delete(CPL_ARRAY, 5, ion_wavel, ion_oscs, ion_damp, par_pos, 
        par_fix);
    cpl_table_delete(par_lim);
    gsl_multifit_fdfsolver_free(solver);
    gsl_matrix_free(covar);
    gsl_vector_free(par_vary);
    return cpl_error_get_code();
}

/**@}*/
