 /* $Id: espda_line.c 235675 2019-02-06 12:05:46Z gcupani $
 *
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Author: gcupani $
 * $Date: 2019-02-06 12:05:46 +0000 (Wed, 06 Feb 2019) $
 * $Revision: 235675 $
 * $Name:  $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include "espda_line.h"

/*----------------------------------------------------------------------------*/
/**
  @defgroup   espda_line  Lists of lines
 
  A list of lines (LINE) is a table with information about absorption lines. 
  Three types of lists are used: basic line list (LINE_BASIC), line list 
  prepared for Voigt-profile fitting (LINE_VOIGT), and line list with 
  information about the absorption systems (LINE_SYST). 
  
  The columns of LINE_BASIC are:
  
  -# @a PEAK: peak value of flux density [erg cm^-2 s^-1 nm^-1];
  -# @a WAVEL: wavelength [nm].
  
  LINE_VOIGT has all all the columns of LINE_BASIC plus the following additional
  columns:
  
  -# @a WAVELERR: error on wavelength [nm]; 
  -# @a WAVELMIN: minimum of the fitted wavelength range [nm];
  -# @a WAVELMAX: maximum of the fitted wavelength range [nm];
  -# @a LINEID: ID;
  -# @a MULTID: ID of the multiplet;
  -# @a GRPID: ID of the group;
  -# @a LINEZ: redshift;
  -# @a LINEZERR: error on redshift;
  -# @a LINEZMIN: minimum redshift for fitting;
  -# @a LINEZMAX: maximum redshift for fitting;
  -# @a LINEZCON: constraint on redshift for fitting;
  -# @a COLDEN: column density [cm^-2, logarithm];
  -# @a COLDENERR: error on column density [cm^-2, logarithm];
  -# @a COLDENMIN: minimum column density for fitting [cm^-2, logarithm];
  -# @a COLDENMAX: maximum column density for fitting [cm^-2, logarithm];
  -# @a COLDENCON: constraint on column density for fitting;
  -# @a THERB: thermal broadening [km s^-1];
  -# @a THERBERR: error on thermal broadening [km s^-1];
  -# @a THERBMIN: minimum thermal broadening for fitting [km s^-1];
  -# @a THERBMAX: maximum thermal broadening for fitting [km s^-1];
  -# @a THERBCON:constraint on thermal broadening for fitting;
  -# @a TURBB: turbulence broadening [km s^-1];
  -# @a TURBBERR: error on turbulence broadening [km s^-1];
  -# @a TURBBMIN: minimum turbulence broadening for fitting [km s^-1];
  -# @a TURBBMAX: maximum turbulence broadening for fitting [km s^-1];
  -# @a TURBBCON: constraint on turbulence broadening for fitting;
  -# @a USE: flag to use the line in fitting;
  -# @a CONT: fitted flux density of the emission continuum  [erg cm^-2 s^-1 nm^-1];
  -# @a NORM: flux density normalized to the emission continuum (@a PEAK/
     @a CONT). 
  
  LINE_SYST has all the columns of LINE_VOIGT plus the following additional
  columns:
    
  -# @a SYSTID: system ID;
  -# @a ORDER: sort order. 
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief   Add lines to a line list.
  @param   line  A list of lines (LINE_BASIC; may be NULL).
  @param   add   Information about the lines to be added.
  @return  CPL error code.
   
  This function adds manually-selected absorption lines to a line list @a line.
  The information about the lines to be added is passed through a string 
  parameter @a add consisting in a comma-separated list of wavelength-peak 
  values, which are appended to the @a WAVEL and @a PEAK columns of @a line.
  
  If @a line or @a add are null and the verbosity level is greater than 0, a 
  warning is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_line_add(cpl_table *line, const char *add) {
    extern int v_level;
    espda_util_warning(line == NULL, "list of lines not provided.");

    char  *add_repl = (char *)malloc(sizeof(char) * LEN_STRING); 
    espda_dfs_ensure(espda_param_replace(add, "-", ",", add_repl));
    cpl_array *add_x = espda_param_extract(add_repl, ",", 0, 2, 
        CPL_TYPE_DOUBLE);
    cpl_array *add_y = espda_param_extract(add_repl, ",", 1, 2, 
        CPL_TYPE_DOUBLE);

    if (add_x != NULL && add_y != NULL && cpl_array_get_size(add_x) ==
        cpl_array_get_size(add_y)) {
        cpl_size num = cpl_array_get_size(add_x);
        cpl_table *line_add = cpl_table_new(num);
        espda_line_basic(line_add);
        cpl_table_copy_data_double(line_add, "WAVEL", 
            cpl_array_get_data_double(add_x));
        cpl_table_copy_data_double(line_add, "PEAK", 
            cpl_array_get_data_double(add_y));
        if (line != NULL) {
            cpl_table_insert(line, line_add, num);
            espda_print_plurt(line_add, add_plur);
            espda_print_plurt(line,     line_plur);
            espda_print_verb(1, cpl_msg_info(cpl_func, " %i line%s added, %i "
                "line%s in total!",  (int)num, add_plur, 
                (int)cpl_table_get_nrow(line), line_plur));
        }
        else {
            line = cpl_table_duplicate(line_add);
        }
        cpl_table_delete(line_add);                                 
        espda_struct_sort(line, 1, "WAVEL", FALSE);
    }
    else {
        espda_print_verb(1, cpl_msg_info(cpl_func, "information about lines to "
                         "be added not provided."));
    }
    espda_struct_delete(CPL_ARRAY, 2, add_x, add_y);
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Clean duplicates from a line list.
  @param   line      A list of lines (LINE_BASIC)
  @param   vel_step  The velocity step to define duplicates [m s^-1].
  @return  CPL error code.
   
  This function finds duplicate entries in a line list @a line and remove them.
  Duplicates are defined as lines separated by less than 1.01 times the velocity 
  step @a vel_step.

  If @a line is null, an error is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_line_clean(cpl_table *line, double vel_step) {
    extern int v_level;
    espda_util_error(line == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    cpl_size i_comp = 0;
    cpl_size num = 0;
    espda_struct_sort(line, 1, "WAVEL", FALSE);
    for (int i = 0; i < cpl_table_get_nrow(line) - 1; i++) {
        double resol = 1.01 * vel_step / LIGHTSPEED *
            cpl_table_get_double(line, "WAVEL", i + 1, NULL);
        if (cpl_table_get_double(line, "WAVEL", i + 1,  NULL) -
            cpl_table_get_double(line, "WAVEL", i, NULL) < resol) {
            if (cpl_table_get_double(line, "PEAK", i + 1, NULL) <
                cpl_table_get_double(line, "PEAK", i_comp, NULL)) {
                cpl_table_set_invalid(line, "PEAK", i_comp);
                i_comp = i + 1;
            }
            else {
                cpl_table_set_invalid(line, "PEAK", i + 1);
            }
            num++;
        } 
        else {
            i_comp = i + 1;
        }
    }
    cpl_table_erase_invalid(line);
    espda_print_plurn(num,  num_plur);
    espda_print_plurt(line, line_plur);
    espda_print_verb(1, cpl_msg_info(cpl_func, " %i duplicate%s cleaned, "
        "%i line%s remaining!", (int)num, num_plur, 
        (int)cpl_table_get_nrow(line), line_plur));
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Detect absorption lines.
  @param    spec     A spectrum (SPEC_BASIC).
  @param    width    Width of the spectrum chunks [nm].
  @param    overlap  Overlap of the specturm chunks.
  @param    thres    Threshold for line detection.
  @return   A list of lines (LINE_BASIC).
   
  This function splits a spectrum @a spec into chunks, checks the flux
  statistics in chunks and detects an absorption line whenever the average flux
  in a chunk is significantly lower than the median flux. Chunks have a width 
  @a width and overlap each other by a fraction @a overlap of @a width. The
  difference between the mean and the median is considered significant if it 
  exceeds a threshold @a thres. The line is placed at the wavelength of lowest 
  flux.
    
  The number of chunks is obtained as @f[
    N_\mathrm{chunks}=\frac{\lambda_\mathrm{max}-\lambda_\mathrm{min}}
                           {\mathit{width}\times(1-\mathit{overlap})},
  @f]
  where @f$ \lambda_\mathrm{min} @f$ and @f$ \lambda_\mathrm{max} @f$ are the 
  minimum and maximum wavelength of the spectrum. The range of the @f$ i @f$-th 
  chunk is defined ad @f[
    C_i = \left[\lambda_\mathrm{min} + 
                (i-1)\times\mathit{width}\times\mathit{overlap},
                \lambda_\mathrm{min} +  
                 i   \times\mathit{width}\times\mathit{overlap}\right].
  @f]

  If @a spec is null, an error is issued.
 */
/*----------------------------------------------------------------------------*/

/*
cpl_error_code espda_line_detect_old(cpl_table *spec, double width, double overlap,
    double thres, cpl_table **line) {
    cpl_errorstate  prestate = cpl_errorstate_get();
    extern int v_level;
    espda_util_error(spec == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    if (spec != NULL) {
        espda_print_verb(1, cpl_msg_info(cpl_func, "Detect lines in "
            "chunks..."));
        cpl_size   line_num = 0;
        cpl_table *line_pre = cpl_table_new(1000);
        espda_line_basic(line_pre);
        //espda_dfs_ensure(espda_struct_line(&line_pre, 1000));
        
        //cpl_size   chunk_num = cpl_table_get_nrow(spec) - width - 1;
        
        espda_struct_sort_old(spec, "WAVEL", FALSE);
        double wavel_min = cpl_table_get_double(spec, "WAVEL", 0, NULL);
        double wavel_max = cpl_table_get_double(spec, "WAVEL", 
            cpl_table_get_nrow(spec) - 1, NULL);
        cpl_size   chunk_num = (int)ceil((wavel_max - wavel_min) / 
            (width * (1 - overlap)));
        //double chunk_min;
        double chunk_max = wavel_min + width * overlap;
            
        for (int i = 0; i < chunk_num; i++) {
        
            //cpl_table *chunk = NULL;
            //espda_struct_extract_old(spec, chunk, NULL, "WINDOW", i, width);
            
            double chunk_min = chunk_max - width * overlap;
            chunk_max = chunk_min + width;
            cpl_table *temp = NULL, *chunk = NULL;
            espda_struct_check(spec, espda_struct_extract_old(spec, temp, "WAVEL", 
                CPL_TYPE_DOUBLE, CPL_NOT_LESS_THAN, chunk_min));
            espda_struct_check(temp, espda_struct_extract_old(temp, chunk, "WAVEL", 
                CPL_TYPE_DOUBLE, CPL_LESS_THAN,     chunk_max));
            espda_struct_check(chunk, espda_struct_sort_old(chunk, "FLUX", FALSE));
            
            int    med_row  = floor((int)cpl_table_get_nrow(chunk) / 2.0);
            double flux_ave = cpl_table_get_column_mean(chunk, "FLUX");
            double flux_med = 
                (cpl_table_get_double(chunk, "FLUX", med_row,     NULL) +
                 cpl_table_get_double(chunk, "FLUX", med_row + 1, NULL)) 
                 / 2.0;
            if ((flux_med - flux_ave) / flux_med > thres) {
                cpl_table_set_double(line_pre, "WAVEL", line_num, 
                    cpl_table_get_double(chunk, "WAVEL", 0, NULL));
                cpl_table_set_double(line_pre, "PEAK",  line_num, 
                    cpl_table_get_double(chunk, "FLUX",  0, NULL));
                line_num++;
            }
        }
        if (line_num > 0) {
            cpl_table *line_red = NULL;
            //espda_struct_extract_old(line_pre, line_red, NULL, "WINDOW", 0, 
            //    line_num);
            espda_struct_check(line_pre, espda_struct_extract_old_2(line_pre, 
                line_red, 0, line_num));
            *line = cpl_table_duplicate(line_red);
            cpl_table_delete(line_red);
            //espda_print_plur(line, plur, "TABLE");
            espda_print_plurt(*line, plur);
            espda_print_verb(1, cpl_msg_info(cpl_func, " %i line%s detected!", 
                (int)line_num, plur));
        }
        else {
            espda_print_verb(1, cpl_msg_info(cpl_func, " 0 lines detected!"));            
        }
        cpl_table_delete(line_pre);
    }    
                               
    if (!cpl_errorstate_is_equal(prestate)) {
        return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
               "Unable to detect absorption lines.");
    }
    return CPL_ERROR_NONE;
}
*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Detect absorption lines.
  @param    spec     A spectrum (SPEC_BASIC).
  @param    kappa    Value to define the selection threshold.
  @return   A list of lines (LINE_BASIC).
   
  This function finds the local maxima and minima of the flux density in a 
  spectrum @a spec and selects as absorption lines the minimum values that 
  differ from (at least one of) the adjacent maximum values more than a given 
  threshold. The wavelengths and flux densities of the minima are taken as
  wavelengths and peak values of the lines.
  
  A minimum is selected as an absorption line if @f[
    f_\mathrm{max}-f_\mathrm{min}>\mathit{kappa}\times\sigma_\mathrm{min},
  @f]
  where @f$ f @fs, @f$ \sigma @f are taken from columns @a FLUX, @a FLUXERR 
  of @a spec, respectively, and the subscripts @f$ \mathrm{min} @f$, 
  @f$ \mathrm{max} @f$ refer to the minimum and to either one of the adjacent 
  maxima.

  If @a spec is null, an error is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_table *espda_line_detect(cpl_table *spec, cpl_table *ref, double kappa) {
    extern int v_level;
    espda_util_error(spec == NULL, CPL_ERROR_NULL_INPUT, NULL);
    int incr_flag = 0;
    int decr_flag = 0;
    cpl_size extr_num = 0;
    cpl_vector *wavel_extr   = cpl_vector_new(cpl_table_get_nrow(spec));
    cpl_vector *flux_extr    = cpl_vector_new(cpl_table_get_nrow(spec));
    cpl_vector *fluxerr_extr = cpl_vector_new(cpl_table_get_nrow(spec));
    cpl_vector *cont_extr    = cpl_vector_new(cpl_table_get_nrow(spec));
    for (int i = 1; i < cpl_table_get_nrow(spec); i++) {
        double wavel    = cpl_table_get_double(spec, "WAVEL",   i - 1, NULL);
        double flux_old = cpl_table_get_double(spec, "FLUX",    i - 1, NULL);
        double flux_new = cpl_table_get_double(spec, "FLUX",    i,     NULL);
        double fluxerr_old = cpl_table_get_double(spec, "FLUXERR", i - 1, NULL);
        //double fluxerr_new = cpl_table_get_double(spec, "FLUXERR", i, NULL);
        double cont = 0.0;

        if (cpl_table_has_column(spec, "CONT")) {
            cont = cpl_table_get_double(spec, "CONT", i - 1, NULL);
        }
        if (flux_new > flux_old) {
            if (incr_flag == 0) {
                cpl_vector_set(wavel_extr,   extr_num, wavel);
                cpl_vector_set(flux_extr,    extr_num, flux_old);
                cpl_vector_set(fluxerr_extr, extr_num, fluxerr_old);
                cpl_vector_set(cont_extr,    extr_num, cont);
                incr_flag = 1;  
                extr_num++;
            }
            decr_flag = 0;
        }
        else if (flux_new < flux_old) {
            if (decr_flag == 0) {
                cpl_vector_set(wavel_extr,   extr_num, wavel);
                cpl_vector_set(flux_extr,    extr_num, flux_old);
                cpl_vector_set(fluxerr_extr, extr_num, fluxerr_old);
                cpl_vector_set(cont_extr,    extr_num, cont);
                decr_flag = 1;
                extr_num++;  
            }
            incr_flag = 0;
        }
    }
    cpl_table *line = NULL;
    if (extr_num > 0) {
        cpl_vector_set_size(wavel_extr,   extr_num);
        cpl_vector_set_size(flux_extr,    extr_num);
        cpl_vector_set_size(fluxerr_extr, extr_num);
        cpl_vector_set_size(cont_extr,    extr_num);
        line = cpl_table_new(extr_num);
        espda_line_basic(line);
        cpl_size line_num = 0;
        if (cpl_table_has_column(spec, "CONT")) {
            espda_struct_initcol(line, "CONT", CPL_TYPE_DOUBLE);
        }

        // Only the difference with one adjacent maximum must be over threshold

/*
        for (int e = 1; e < extr_num; e++) {
            double check = cpl_vector_get(flux_extr, e) - 
                cpl_vector_get(flux_extr, e - 1);
            double thres = cpl_vector_get(fluxerr_extr, e) * kappa;
            if (check > thres) {
                cpl_table_set_double(line, "WAVEL", line_num, 
                    cpl_vector_get(wavel_extr, e - 1));
                cpl_table_set_double(line, "PEAK",  line_num, 
                    cpl_vector_get(flux_extr,  e - 1));
                if (cpl_table_has_column(spec, "CONT")) {
                    cpl_table_set_double(line, "CONT",  line_num, 
                        cpl_vector_get(cont_extr,  e - 1));
                
                }
                //printf("%f %f %f %f %f\n", cpl_vector_get(wavel_extr, e), cpl_vector_get(flux_extr, e), cpl_vector_get(fluxerr_extr, e), kappa, thres);
                line_num++;
            }
            else if (check < -thres) {
                cpl_table_set_double(line, "WAVEL", line_num, 
                    cpl_vector_get(wavel_extr, e));
                cpl_table_set_double(line, "PEAK",  line_num, 
                    cpl_vector_get(flux_extr,  e));
                if (cpl_table_has_column(spec, "CONT")) {
                    cpl_table_set_double(line, "CONT",  line_num, 
                        cpl_vector_get(cont_extr,  e));
                }
                //printf("%f %f %f %f %f\n", cpl_vector_get(wavel_extr, e), cpl_vector_get(flux_extr, e), cpl_vector_get(fluxerr_extr, e), kappa, thres);
                line_num++;
            }
        }
*/
        // The difference with both adjacent maxima must be over threshold

        for (int e = 1; e < extr_num - 1; e++) {
            double check_l = cpl_vector_get(flux_extr, e - 1) - 
                cpl_vector_get(flux_extr, e);
            double check_u = cpl_vector_get(flux_extr, e + 1) - 
                cpl_vector_get(flux_extr, e);
            double thres = (//cpl_vector_get(fluxerr_extr, e)
			    //+
			    cpl_vector_get(fluxerr_extr, e - 1)
			    + cpl_vector_get(fluxerr_extr, e + 1)) / 2.0
			    * kappa;
            if (check_l > thres && check_u > thres) {
	      //if (check_l > thres || check_u > thres) {
                cpl_table_set_double(line, "WAVEL", line_num, 
                    cpl_vector_get(wavel_extr, e));
                cpl_table_set_double(line, "PEAK",  line_num, 
                    cpl_vector_get(flux_extr,  e));
                line_num++;
                if (cpl_table_has_column(spec, "CONT")) {
                    cpl_table_set_double(line, "CONT",  line_num, 
                        cpl_vector_get(cont_extr,  e));
                
                }
            }
        }

        cpl_table_set_size(line, line_num);
        espda_print_plurt(line, plur);
        espda_print_verb(1, cpl_msg_info(cpl_func, " %i line%s detected!", 
            (int)line_num, plur));
    }

    // Shift the wavelength to the closest minima in the reference spectrum
    // Closest minima are found by going downhill from the minima of the
    // smoothed spectrum in both directions; the algorithm continues to go
    // downhill until it reaches an obstacle greater than the local error on
    // flux, times kappa
    for (int l = 0; l < cpl_table_get_nrow(line); l++) {
        double wavel = cpl_table_get_double(line, "WAVEL", l, NULL);
        //double peak = cpl_table_get_double(line, "PEAK", l, NULL);
	cpl_table_duplicate_column(ref, "WAVEL_ABS", ref, "WAVEL");
	cpl_table_subtract_scalar(ref, "WAVEL_ABS", wavel);
	cpl_table_abs_column(ref, "WAVEL_ABS");
	cpl_size minpos;
	cpl_table_get_column_minpos(ref, "WAVEL_ABS", &minpos);
	/*
	int minpos_m = minpos;
	int minpos_ref_m = minpos;
	double wavel_min_m = wavel;
	double peak_min_m = peak;
	*/
	//double peak_min = peak;
	/*
	  while (minpos_m > 1 &&
	       cpl_table_get_double(ref, "FLUX", minpos_m - 1, NULL) <
	       cpl_table_get_double(ref, "FLUX", minpos_m, NULL) +
	       kappa * cpl_table_get_double(ref, "FLUXERR", minpos_m, NULL)) {
	    //3.0 * cpl_table_get_double(ref, "FLUXERR", minpos_m, NULL)) {
	  //wavel_min_p = cpl_table_get_double(ref, "WAVEL", minpos_m - 1, NULL);
	  //peak_min = cpl_table_get_double(ref, "FLUX", minpos_m - 1, NULL);	
	    minpos_m--;
	    if (cpl_table_get_double(ref, "FLUX", minpos_m, NULL) < peak_min_m || 1!=1){
 	        minpos_ref_m = minpos_m;
		peak_min_m = cpl_table_get_double(ref, "FLUX", minpos_m, NULL);
	    }
	}
	*/
	/*
	int minpos_p = minpos;
	int minpos_ref_p = minpos;
	double wavel_min_p = wavel;
	double peak_min_p = peak;
	*/
	/*
	  while (minpos_p < cpl_table_get_nrow(ref) - 1 &&
	       cpl_table_get_double(ref, "FLUX", minpos_p + 1, NULL) <
	       cpl_table_get_double(ref, "FLUX", minpos_p, NULL) +
	       kappa * cpl_table_get_double(ref, "FLUXERR", minpos_p, NULL)) {
	    //3.0 * cpl_table_get_double(ref, "FLUXERR", minpos_p, NULL)) {
	  //wavel_min = cpl_table_get_double(ref, "WAVEL", minpos_p + 1, NULL);
	  //peak_min = cpl_table_get_double(ref, "FLUX", minpos_p + 1, NULL);	
	    minpos_p++;
	    if (cpl_table_get_double(ref, "FLUX", minpos_p, NULL) < peak_min_p || 1!=1){
 	        minpos_ref_p = minpos_p;
		peak_min_p = cpl_table_get_double(ref, "FLUX", minpos_p, NULL);
	    }
        }
	*/
	/*
	if (cpl_table_get_double(ref, "FLUX", minpos_ref_m, NULL) <
	    cpl_table_get_double(ref, "FLUX", minpos_ref_p, NULL)) {
	    cpl_table_set_double(line, "WAVEL", l, cpl_table_get_double(
		ref, "WAVEL", minpos_ref_m, NULL));
	    cpl_table_set_double(line, "PEAK", l, cpl_table_get_double(
		ref, "FLUX", minpos_ref_m, NULL));
	}
	else {
	    cpl_table_set_double(line, "WAVEL", l, cpl_table_get_double(
		ref, "WAVEL", minpos_ref_p, NULL));
	    cpl_table_set_double(line, "PEAK", l, cpl_table_get_double(
		ref, "FLUX", minpos_ref_p, NULL));
	}
	*/
	cpl_table_set_double(line, "WAVEL", l, cpl_table_get_double(
		ref, "WAVEL", minpos, NULL));
	cpl_table_set_double(line, "PEAK", l, cpl_table_get_double(
		ref, "FLUX", minpos, NULL));
	    //cpl_table_set_double(line, "WAVEL", l, wavel_min_p);
	//    cpl_table_set_double(line, "PEAK", l, peak_min_p);
	cpl_table_erase_column(ref, "WAVEL_ABS");
    }
    
    espda_struct_delete(CPL_VECTOR, 4, wavel_extr, flux_extr, fluxerr_extr,
                        cont_extr); 
    return line;
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Fill columns with Voigt parameters.
  @param   line_in    A list of lines (LINE_BASIC).
  @param   hwidth     Half-width of the allowed wavelength range [nm].
  @param   z_em       The emission redshift of the quasar.
  @param   par_range  Ranges for the Voigt parameters.
  @param   line_out   A new filled list of lines (LINE_VOIGT).
  @return  CPL error code.
  
  This function takes a line list @a line_in and creates a duplicate line list
  @a line_out filled with guess Voigt parameters for fitting. The additional 
  columns in @a line_out are filled as follows:
  
  -# @a LINEID: @p Ly_a for lines in the Ly-alpha forest, @p CIV_1548 for 
     lines in the CIV forest, and MgII_2803 for lines with greater wavelength 
     (@a z_em is used to compute the regions);
  -# @a WAVELMIN: @a WAVEL - @a hwidth;
  -# @a WAVELMAX: @a WAVEL + @a hwidth;
  -# @a LINEZCON: @p V;
  -# @a COLDEN: modeled on peak value for @p Ly_a lines, 13.0 otherwise;
  -# @a COLDENMIN: @a par_range[1];
  -# @a COLDENMAX: @a par_range[2];
  -# @a COLDENCON: @p V;
  -# @a THERB: 20.0 for @p Ly_a lines, 10.0 otherwise;
  -# @a THERBMIN: @a par_range[3];
  -# @a THERBMAX: @a par_range[4];
  -# @a THERBCON: @p V;
  -# @a TURBB: 0.0;
  -# @a TURBBMIN: @a par_range[5];
  -# @a TURBBMAX: @a par_range[6];
  -# @a TURBBCON: @p F;
  -# @a USE: @p Y.
  
  Columns @a LINEZ, @a LINEZMIN, and @a LINEZMAX are filled by function 
  @p espda_line_group instead.

  For details on the modeling of the column density, see @p espda_model_colden.
  
  If @a line_in is NULL, an error is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_line_fill(cpl_table *line_in, double hwidth, double zem,
    double *par_range_value, cpl_table **line_out) {
    extern int v_level;
    espda_util_error(line_in == NULL, CPL_ERROR_NULL_INPUT, 
        cpl_error_get_code());
    //cpl_errorstate  prestate = cpl_errorstate_get();
    cpl_size   line_size = cpl_table_get_nrow(line_in);
    *line_out = cpl_table_new(line_size);
    if (!cpl_table_has_column(line_in, "SYSTID")) {
        espda_line_voigt(*line_out);
    }
    else {
        espda_line_syst(*line_out);
    }
    for (int i = 0; i < line_size; i++) {
        double  wavel  = cpl_table_get_double(line_in, "WAVEL", i, NULL); 
        double  peak   = cpl_table_get_double(line_in, "PEAK",  i, NULL); 
        double  cont   = cpl_table_get_double(line_in, "CONT",  i, NULL); 
        char   *lineid = malloc(sizeof(char) * LEN_STRING);
        char   *multid = malloc(sizeof(char) * LEN_STRING);
        char   *systid = malloc(sizeof(char) * LEN_STRING);
        strcpy(lineid, "Ly_a");
        strcpy(multid, INV_STRING);
        strcpy(systid, INV_STRING);
        double  colden = espda_model_colden(peak, cont);
        double  therb  = 20.0;
	//printf("%f %f\n", peak, colden);
        if (wavel > WAVEL_LY_A * (1.0 + zem)) {
            if (wavel <= WAVEL_CIV_1550 * (1.0 + zem)) { 
                strcpy(lineid, "CIV_1548");
                strcpy(multid, "CIV");
            }
            else {
                strcpy(lineid, "MgII_2803");
                strcpy(multid, "MgII");
            }
            colden = colden - 1.0;
            therb  = 5.0;
        }
        hwidth = 2.35482 * therb / LIGHTSPEED * wavel;  // FWHM
	
	//printf("%f\n", colden);
        if (cpl_table_has_column(line_in, "LINEID")) {
            if (cpl_table_get_string(line_in, "LINEID",  i) != NULL) {
                strcpy(lineid, cpl_table_get_string(line_in, "LINEID",  i));
            }
        }
        if (cpl_table_has_column(line_in, "MULTID")) {
            if (cpl_table_get_string(line_in, "MULTID",  i) != NULL) {
                strcpy(multid, cpl_table_get_string(line_in, "MULTID",  i));
            }
        }
        if (cpl_table_has_column(line_in, "SYSTID")) {
            if (cpl_table_get_string(line_in, "SYSTID",  i) != NULL) {
                strcpy(systid, cpl_table_get_string(line_in, "SYSTID",  i));
            }
        }
        cpl_table_set_double(*line_out, "WAVEL",  i, wavel);
        cpl_table_set_double(*line_out, "PEAK",   i, peak);
        cpl_table_set_double(*line_out, "CONT",   i, cont);
        cpl_table_set_string(*line_out, "LINEID", i, lineid);
        cpl_table_set_string(*line_out, "MULTID", i, multid);
        if (cpl_table_has_column(*line_out, "SYSTID")) {
            cpl_table_set_string(*line_out, "SYSTID", i, systid);
        }
        cpl_table_set_double(*line_out, "COLDEN", i, colden);
        cpl_table_set_double(*line_out, "THERB",  i, therb);
        if (cpl_table_has_column(line_in, "ORDER")) {
            espda_struct_copy(*line_out, "ORDER", i, line_in, "ORDER", i);
        }
	free(lineid);
	free(multid);
        free(systid);
    }
    cpl_table_fill_column_window_double(*line_out, "WAVELMIN",  0, line_size, 
        -hwidth);
    cpl_table_add_columns(*line_out, "WAVELMIN",  "WAVEL");
    cpl_table_fill_column_window_double(*line_out, "WAVELMAX",  0, line_size,  
        hwidth);
    cpl_table_add_columns(*line_out, "WAVELMAX",  "WAVEL");
    cpl_table_fill_column_window_string(*line_out, "LINEZCON",  0, line_size, 
        "V");
    cpl_table_fill_column_window_double(*line_out, "COLDENMIN", 0, line_size, 
        par_range_value[1]);
    cpl_table_fill_column_window_double(*line_out, "COLDENMAX", 0, line_size, 
        par_range_value[2]);
    cpl_table_fill_column_window_string(*line_out, "COLDENCON", 0, line_size, 
        "V");
    cpl_table_fill_column_window_double(*line_out, "THERBMIN",  0, line_size, 
        par_range_value[3]);
    cpl_table_fill_column_window_double(*line_out, "THERBMAX",  0, line_size, 
        par_range_value[4]);
    cpl_table_fill_column_window_string(*line_out, "THERBCON",  0, line_size, 
        "V");
    cpl_table_fill_column_window_double(*line_out, "TURBB",     0, line_size, 
        0.0);
    cpl_table_fill_column_window_double(*line_out, "TURBBMIN",  0, line_size, 
        par_range_value[5]); 
    cpl_table_fill_column_window_double(*line_out, "TURBBMAX",  0, line_size, 
        par_range_value[6]); 
    cpl_table_fill_column_window_string(*line_out, "TURBBCON",  0, line_size, 
        "F");
    cpl_table_fill_column_window_string(*line_out, "USE",       0, line_size, 
        "Y");
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Define groups in a line list.
  @param   line       A list of lines (LINE_VOIGT).
  @param   ion        A list of ionic transitions (ION).
  @param   par_range  Ranges for the Voigt parameters.
  @return  CPL error code.
  
  This function scans a list of lines @a line and groups the lines that must be
  fitted together. Lines are grouped if their wavelength ranges available for
  fitting (defined by columns @a WAVELMIN and @a WAVELMAX) overlap, or if 
  specific constraints on their Voigt parameters (defined in columns 
  @a LINEZCON, @a COLDENCON, @a THERBCON, and @a TURBBCON) have been set. Groups
  of lines are defined by incremental integer values in column @a GRPID.

  The function also fills the following columns of @a line:

  -# @a LINEZ: computed using the reference list of ionic transitions @a ion;
  -# @a LINEZMIN: @a LINEZ - @a par_range[0];
  -# @a LINEZMAX: @a LINEZ + @a par_range[0];
  
  If @a line is NULL, an error is issued.
  
  N.B. It could be reasonable to move the part in which columns @a LINEZ, 
       @a LINEZMIN, and @a LINEZMAX are filled to function @p espda_line_fill.
       However, this affects the continuum fitting because the redshift range 
       for fitting line in the Ly-alpha forest would be fixed once and for all
       and not updated at each iteration. The results were not satisfying, so 
       the code has been reverted (at least temporary) to have the redshift 
       range defined by this function.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_line_group(cpl_table *line, cpl_table *ion,
    double *par_range_value) {
    //cpl_errorstate prestate = cpl_errorstate_get();
    extern int v_level;
    espda_util_error(line == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    espda_util_error(line == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());

    cpl_size line_size = cpl_table_get_nrow(line);
    int grp_id = 0;
     
    for (int i = 0; i < line_size; i++) {

        // Group lines
        double      wavelmin_i   = cpl_table_get_double(line, "WAVELMIN",  i,
                                                        NULL);
        double      wavelmax_i   = cpl_table_get_double(line, "WAVELMAX",  i,
                                                        NULL);
        const char *linez_con_i  = cpl_table_get_string(line, "LINEZCON",  i);
        const char *colden_con_i = cpl_table_get_string(line, "COLDENCON", i);
        const char *therb_con_i  = cpl_table_get_string(line, "THERBCON",  i);
        const char *turbb_con_i  = cpl_table_get_string(line, "TURBBCON",  i);
        int     grp_id_i         = cpl_table_get_int   (line, "GRPID",     i,
                                                        NULL);
        // For quasars, compute line redshifts   
        const char *lineid_sel = cpl_table_get_string(line, "LINEID", i);
        cpl_table_unselect_all(ion);
        
        // CPL_EQUAL_TO cannot be used, otherwise the string are treated as
        // regular expressions and the asterisk are not taken literally
        cpl_table_or_selected_string (ion, "LINEID", CPL_NOT_GREATER_THAN, 
                                      lineid_sel);
        cpl_table_and_selected_string(ion, "LINEID", CPL_NOT_LESS_THAN, 
                                      lineid_sel);
        
        cpl_table *ion_sel = cpl_table_extract_selected(ion);
        if (cpl_table_get_nrow(ion_sel) > 1) {
            cpl_msg_warning(cpl_func, "Reference line list has more than "
                            "one row with LINEID %s", lineid_sel);
        }
        double wavel_ref = cpl_table_get_double(ion_sel, "WAVEL",  0, NULL);
        double linez     = cpl_table_get_double(line, "WAVEL",    i, NULL) 
                           / wavel_ref - 1;
        double linez_min = linez - par_range_value[0];
        double linez_max = linez + par_range_value[0];
        cpl_table_set_double(line, "LINEZ",    i, linez);
        cpl_table_set_double(line, "LINEZMIN", i, linez_min);
        cpl_table_set_double(line, "LINEZMAX", i, linez_max);
        cpl_table_delete(ion_sel);

        int grp_flag = 0;
        for (int ii = 0; ii < line_size; ii++) {
            double      wavelmin_ii   = cpl_table_get_double(line, 
                                        "WAVELMIN",    ii, NULL);
            double      wavelmax_ii   = cpl_table_get_double(line, 
                                        "WAVELMAX",    ii, NULL);
            const char *linez_con_ii  = cpl_table_get_string(line, 
                                        "LINEZCON",    ii);
            const char *colden_con_ii = cpl_table_get_string(line,
                                        "COLDENCON",   ii);
            const char *therb_con_ii  = cpl_table_get_string(line, 
                                        "THERBCON",    ii);
            const char *turbb_con_ii  = cpl_table_get_string(line, 
                                        "TURBBCON",    ii);
            if (( strcmp(linez_con_i,  "V")                 &&
                  strcmp(linez_con_i,  "F")                 &&
                 !strcmp(linez_con_i,  linez_con_ii))       ||
                ( strcmp(colden_con_i, "V")                 &&
                  strcmp(colden_con_i, "F")                 &&
                 !strcmp(colden_con_i, colden_con_ii))      ||
                ( strcmp(therb_con_i,  "V")                 &&
                  strcmp(therb_con_i,  "F")                 &&
                 !strcmp(therb_con_i,  therb_con_ii))       ||
                ( strcmp(turbb_con_i,  "V")                 &&
                  strcmp(turbb_con_i,  "F")                 &&
                 !strcmp(turbb_con_i,  turbb_con_ii))       ||
                (wavelmin_ii < wavelmin_i  && 
                 wavelmax_ii > wavelmin_i) ||
                (wavelmin_ii < wavelmax_i  &&
                 wavelmax_ii > wavelmin_i)) {
                if (grp_id_i == INV_INT) {                       
                    cpl_table_set_int(line, "GRPID", ii, grp_id);
                    grp_flag = 1; 
                }
                else {
                    cpl_table_set_int(line, "GRPID", ii, grp_id_i);
                    grp_flag = 0; 
                }
            }
             
        }
        if (grp_flag != 0) {
            grp_id++;
        }
    }
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Identify lines in a stellar spectrum.
  @param   line      A list of lines (LINE_BASIC).
  @param   ion       A list of ionic transitions (ION).
  @param   wdiff     Wavelength difference for matching lines [nm].
  @param   line_id   The list of identified lines (LINE_BASIC).
  @param   line_rej  The list of non-identified lines (LINE_BASIC).
  @param   id_num    Number of identified lines.
  @param   rej_num   Number of non-identified lines.
  @return  CPL error code
   
  This function takes a list of absorption lines @a line from a stellar 
  spectrum, compares it with a list of ionic transitions @a ion and find the 
  correspondences within a given wavelength difference @a wdiff. If a 
  correspondence is found, the line has been identified. A list of identified 
  lines @a line_id and a list of non-identified @a line_rej lines are produced. 
  Lines IDs and wavelengths are copied from the input list to the list of 
  identified lines. The number of identified lines @a id_num and the number of 
  non-identified lines @a rej_num are also computed.
  
  If @a line or @a ion are null, an error is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_line_iden(cpl_table *line, cpl_table *ion, double wdiff, 
    cpl_table **line_id, cpl_table **line_rej, cpl_size *out_num,
    cpl_size *rej_num) {
    extern int v_level;
    espda_util_error(line == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    espda_util_error(ion  == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    *line_rej = cpl_table_duplicate(line);

/*
 * Reference list is scanned. For each line, correspondence within the allowed
 * wavelength difference are searched. Line wavelengths and IDs are loaded. 
 * Found lines are flagged in the duplicate table.
 * 
 * NB: Since the provisional reference list does not contain line IDs, this 
 *     information is not loaded at present.
 */
 
    espda_print_verb(1, cpl_msg_info(cpl_func, "Identify lines..."));
    cpl_size   in_num  = cpl_table_get_nrow(line);
    cpl_size   ref_num = cpl_table_get_nrow(ion);
    cpl_array *wavel   = cpl_array_new(ref_num, CPL_TYPE_DOUBLE);
    //cpl_array *lineid  = cpl_array_new(ref_num, CPL_TYPE_STRING);

    int k = 0, inv_count = 0;
    for (int i = 0; i < ref_num; i++) {

        // Look for correspondences
        double wmin = cpl_table_get_double(ion, "WAVEL", i, NULL) - wdiff;
        double wmax = cpl_table_get_double(ion, "WAVEL", i, NULL) + wdiff;
        cpl_table *temp = NULL, *line_pre = NULL;
        espda_struct_extract_old(line, temp,  "WAVEL", CPL_TYPE_DOUBLE, 
            CPL_GREATER_THAN, wmin);
        espda_util_safe(temp, espda_struct_extract_old(temp, line_pre, "WAVEL", 
	    CPL_TYPE_DOUBLE, CPL_LESS_THAN,    wmax));
	cpl_table_delete(temp);
        cpl_size   line_pre_num = 0;  
        espda_util_safe(line_pre, line_pre_num = cpl_table_get_nrow(line_pre));

        // Load wavelengths (and line IDs, in the future)
        for (int j = 0; j < line_pre_num; j++) {
            cpl_array_set_double(wavel, k, cpl_table_get_double(line_pre, 
                "WAVEL",  j, NULL));
            k++;
        }
        cpl_table_delete(line_pre);

        // Flag found lines
        cpl_table_unselect_all(*line_rej);
        cpl_table_or_selected_double (*line_rej, "WAVEL", CPL_GREATER_THAN, 
            wmin);
        cpl_table_and_selected_double(*line_rej, "WAVEL", CPL_LESS_THAN, wmax);
        for (int j = 0; j < in_num; j++) {
            if (cpl_table_is_selected(*line_rej, j)) {
                inv_count++;
                cpl_table_set_invalid(*line_rej, "WAVEL", j);
            }
        }
    }
    cpl_array_set_size(wavel, k);
    *out_num = cpl_array_get_size(wavel);
    espda_print_plurn(*out_num, out_plur);
    espda_print_verb(1, cpl_msg_info(cpl_func, " %i line%s identified!", 
        (int)*out_num, out_plur));

/*
 * A list of identified lines is created. Non-identified lines are erased from
 * the duplicate table.
 */

    *line_id = cpl_table_new(*out_num);
    espda_line_basic(*line_id);
    for (int l = 0; l < *out_num; l++) {
        cpl_table_set_double(*line_id, "WAVEL",  l, cpl_array_get_double(wavel, 
            l, NULL));
    }
    cpl_array_delete(wavel);
    //cpl_array_delete(lineid);

    // Erase non-identified lines
    cpl_table_erase_invalid(*line_rej);
    if (inv_count == in_num) {
        *rej_num = 0;
    }
    else {
        *rej_num = cpl_table_get_nrow(*line_rej);
    }
    espda_print_plurn(*rej_num, rej_plur);
    espda_print_verb(1, cpl_msg_info(cpl_func, " %i line%s not "
        "identified...", (int)*rej_num, rej_plur));
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Extract information about ionic transitions.
  @param   line   A list of lines (LINE_VOIGT).
  @param   ion    A list of ionic transitions (ION).
  @param   ion_wavel  Wavelengths of the ions [nm].
  @param   ion_oscs   Oscillator strengths of the ions.
  @param   ion_damp   Damping parameters of the ions.
  @return  CPL error code.
  
  This function takes a list @a line of identified absorption lines and for each
  line it extracts from a list of ionic transitions @a ion the wavelength 
  @a ion_wavel, the oscillator strength @a ion_oscs, and the damping parameter 
  @a ion_damp of the corresponding ion. These parameters are given in output as
  arrays with the same length of @a line.
 
  If @a line or @ion are NULL, an error is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_line_ion(cpl_table *line, cpl_table *ion,
    cpl_array **ion_wavel, cpl_array **ion_oscs, cpl_array **ion_damp) {
    espda_util_error(line == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    espda_util_error(ion  == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    cpl_size   line_size = cpl_table_get_nrow(line);
    *ion_wavel = cpl_array_new(line_size, CPL_TYPE_DOUBLE);
    *ion_oscs  = cpl_array_new(line_size, CPL_TYPE_DOUBLE);
    *ion_damp  = cpl_array_new(line_size, CPL_TYPE_DOUBLE);
    for (int i = 0; i < (int)line_size; i++) {
        const char *line_id = cpl_table_get_string(line, "LINEID", i);
        cpl_table_unselect_all(ion);
        
        // CPL_EQUAL_TO cannot be used, otherwise the string are treated as
        // regular expressions and the asterisk are not taken literally
        cpl_table_or_selected_string (ion, "LINEID", CPL_NOT_GREATER_THAN, 
                                      line_id);
        cpl_table_and_selected_string(ion, "LINEID", CPL_NOT_LESS_THAN, 
                                      line_id);

        cpl_table *line = cpl_table_extract_selected(ion);
        cpl_array_set_double(*ion_wavel, i, 
            cpl_table_get_double(line, "WAVEL", 0, NULL));
        cpl_array_set_double(*ion_oscs, i, 
            cpl_table_get_double(line, "OSCS",  0, NULL));
        cpl_array_set_double(*ion_damp, i, 
            cpl_table_get_double(line, "DAMP",  0, NULL));
        cpl_table_delete(line);
    }
    return cpl_error_get_code();
}

/**@}*/
