/* $Id: espda_mask.c 215193 2018-01-22 12:00:18Z gcupani $
 *
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Author: gcupani $
 * $Date: 2018-01-22 12:00:18 +0000 (Mon, 22 Jan 2018) $
 * $Revision: 215193 $
 * $Name:  $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>

#include "espda_mask.h"

/*----------------------------------------------------------------------------*/
/**
  @defgroup  espda_mask  Spectral masks
  
  A spectral mask (MASK) is a table containing a list of regions not to be 
  considered for spectral analysis. The columns of MASK are:
  
  -# @a EXP: ID of the spectral exposure to which the mask is to be applied (0 
     for all exposures);
  -# @a MREGSTART: minimum wavelength of the masked region [nm]; 
  -# @a MREGEND: maximum wavelength of the masked region [nm]. 
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief   Apply a spectral mask.
  @param   spec     A spectrum (may be NULL).
  @param   mask     A spectral mask (may be NULL).
  @param   row_num  The number of masked rows.
  @return  CPL_ERROR_NONE iff OK.

  This function applies a spectral mask @a mask to a spectrum @a spec. Within 
  the masked regions the spectrum wavelengths are flagged as invalid. If @a spec
  contains multiple exposures, only those specified in the mask are actually
  masked. The number of masked rows @a row_num is also computed. The application 
  of a mask is additive, i.e. it preserves any other mask previously applied.

  If @a spec or @a mask are NULL and the verbosity level is greater than 0, a 
  warning is issued. 
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_mask_apply(cpl_table *spec, cpl_table *mask,
    cpl_size *row_num) {
    extern int v_level;
    extern char *qual;
    espda_util_warning(spec == NULL, "spectrum not provided.");    
    espda_util_warning(mask == NULL, "mask not provided.");    
    if (spec != NULL && mask != NULL) {
        cpl_size spec_size = cpl_table_get_nrow(spec);
        cpl_size mask_size = cpl_table_get_nrow(mask);
        for (int k = 0; k < (int)mask_size; k++) {
            int    exp       = cpl_table_get_int   (mask, "EXP",       k, NULL);
            double mregstart = cpl_table_get_double(mask, "MREGSTART", k, NULL);
            double mregend   = cpl_table_get_double(mask, "MREGEND",   k, NULL);
            for (int l = 0; l < (int)spec_size; l++) {
            
                // Exposures are selected
                int specid = 0;
                if (cpl_table_has_column(spec, "SPECID")) {
                    specid = cpl_table_get_int(spec, "SPECID", l, NULL);
                }
                double wavel = cpl_table_get_double(spec, "WAVEL",  l, NULL);
                
                // Wavelengths are flagged as invalid in the masked regions of
                // the selected exposures 
                if ((wavel >= mregstart && wavel <= mregend) &&
                    (exp == 0 || specid == exp)) {
                    cpl_table_set_invalid(spec, "WAVEL", l);
                }
            }
        }
    } 
    *row_num = cpl_table_count_invalid(spec, "WAVEL");
    cpl_table_erase_invalid(spec);
    espda_print_verb(1, cpl_msg_info(cpl_func, " %s spectrum: %-6i/%-6i "
        "pixels left.", qual, (int)(cpl_table_get_nrow(spec) - *row_num), 
        (int)cpl_table_get_nrow(spec)));        
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Create a spectral mask.
  @param   exp  The IDs of the exposures (may be NULL).
  @param   lim  The limits of the regions (may be NULL).
  @return  A spectral mask (MASK).
 
  This functions creates a spectral mask from two vectors @a exp and @a lim,
  respectively containing the IDs of the exposures to be masked and the 
  wavelengths limits of the region to be masked. 
  
  If @a exp and @a lim do not have the same even number of elements, an error is
  issued. If @a exp or @a lim are NULL and the verbosity level is greater than 
  0, a warning is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_table *espda_mask_create(cpl_vector *exp,cpl_vector  *lim) {
    extern int v_level;
    espda_util_warning(exp == NULL, "exposure numbers not provided.");    
    espda_util_warning(lim == NULL, "limits of the regions not provided.");    
    cpl_table *mask = NULL;
    if (exp != NULL && lim != NULL) {
        int size_check = cpl_vector_get_size(exp) != cpl_vector_get_size(lim);
        espda_util_error(size_check != 0, CPL_ERROR_INCOMPATIBLE_INPUT, NULL);
        int even_check = cpl_vector_get_size(exp) % 2;
        espda_util_error(even_check != 0, CPL_ERROR_ILLEGAL_INPUT, NULL);
        
        // Vectors are converted into a table
        double *exp_data = cpl_vector_get_data(exp);
        double *lim_data = cpl_vector_get_data(lim);
        cpl_table *pre = cpl_table_new(cpl_vector_get_size(exp));
        cpl_table_new_column(pre, "EXP", CPL_TYPE_INT);
        cpl_table_new_column(pre, "LIM", CPL_TYPE_DOUBLE);
        for (int j = 0; j < (int)cpl_vector_get_size(exp); j++) {
            cpl_table_set_int   (pre, "EXP", j, (int)exp_data[j]);
            cpl_table_set_double(pre, "LIM", j,      lim_data[j]);
        }
        espda_struct_sort(pre, 1, "EXP", TRUE);
        int exp_max = cpl_table_get_int(pre, "EXP", 0, NULL);
        espda_struct_sort(pre, 1, "EXP", FALSE);
        int exp_min = cpl_table_get_int(pre, "EXP", 0, NULL);
 
        mask = cpl_table_new(cpl_vector_get_size(exp));
        cpl_table_new_column(mask, "EXP",       CPL_TYPE_INT);
        cpl_table_new_column(mask, "MREGSTART", CPL_TYPE_DOUBLE);
        cpl_table_new_column(mask, "MREGEND",   CPL_TYPE_DOUBLE);   
        int k = 0;
        for (int i = exp_min; i <= exp_max; i++) {
        
            // Exposures are selected
            cpl_table *temp = espda_struct_extract(pre, 1, "EXP", CPL_EQUAL_TO, 
                i);
            espda_struct_sort(temp, 1, "LIM", FALSE);
            if (cpl_table_get_nrow(temp) % 2 != 0) {
                cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT, 
                    "Exposure %i is not associated with an even number of " 
                    "wavelength limits.", i);        
            }
            else {   
                
                // Data are copied into the mask
                for (int j = 0; j < cpl_table_get_nrow(temp); j++) {
                    if ((j - 1) % 2 == 0) {
                        espda_struct_copy(mask, "EXP",       k, temp, "EXP", 
                            j - 1);
                        espda_struct_copy(mask, "MREGSTART", k, temp, "LIM", 
                            j - 1);
                        espda_struct_copy(mask, "MREGEND",   k, temp, "LIM", 
                            j);
                        k++;
                    }
                }
            }
            cpl_table_delete(temp);
        }
        cpl_table_delete(pre);
        cpl_table_set_size(mask, k);
    }
    return mask;
}

/**@}*/
