/* 
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Id: espda_redsh.c 161538 2014-05-22 13:56:10Z gcupani $
 * $Author: gcupani $
 * $Date: 2014-05-22 15:56:10 +0200 (Thu, 22 May 2014) $
 * $Revision: 161538 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include "espda_redsh.h"

/*----------------------------------------------------------------------------*/
/**
  @defgroup  espda_redsh  Lists of redshifts
  
  A list of redshifts is a special list of lines (LINE_SYST) produced by 
  matching a list of lines with Voigt profiles (LINE_VOIGT) with a list of 
  ionic transitions (ION). It contains all possible redshift values for the 
  lines.
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief   Group redshift values into candidate systems.
  @param   redsh      A list of redshifts (LINE_SYST).
  @param   grp_width  The width of the window used for grouping [nm].
  @param   grp_num    The minimum number of lines in a group.
  @param   syst_num   The number of candidate systems.
  @return  The list of candidate systems (LINE_SYST).
  
  This functions scans a list of redshifts @a redsh, groups the values that 
  match within a given window with width @a grp_width and checks that the number
  of distinct transitions in the group are above a given threshold @a grp_num. 
  If so, the group is considered a candidate absorption systems. A list of 
  @a syst_num candidate systems is produced and its @a SYSTID column is filled
  with the system IDs. 
 
  Lines are sorted by ascending @a WAVEL. A tentative group with @a grp_num 
  lines is created starting from the first line in the list. The average 
  @a LINEZ of this group, @f$ \bar{z} @f$, is tested against the following 
  conditions: 
  @f[
    \left|\frac{1+\bar{z}}{1+z_\mathrm{first}}-1\right| <
      \frac{\mathit{grp\_width}}{\lambda_\mathrm{first}},
  @f]
  @f[
    \left|\frac{1+\bar{z}}{1+z_\mathrm{last}} -1\right| <
      \frac{\mathit{grp\_width}}{\lambda_\mathrm{last}}.
  @f]
  where @f$ \lambda_\mathrm{first} @f$ and @f$ z_\mathrm{first} @f$,
  @f$ \lambda_\mathrm{last} @f$ and @f$ z_\mathrm{last} @f$ are the @a WAVEL
  and @a LINEZ of the first, last line of the group, respectively. If the group
  passes the test, it is regarded as a candidate system. In this case new lines 
  at greater wavelength are added one by one to the "tail" of the system until
  the conditions are met.  
  When the systems fails the test after adding one line it is regarded as 
  complete. A new tentative group of @a grp_num lines is created starting from
  the second line in the list. This new group undergoes the same treatment. If
  the group is selected as a new candidate, it is tested against an additional
  condition, to determine if it is actually independent from the previous one:
  @f[
    \left|\frac{1+\bar{z}_\mathrm{old}}{1+\bar{z}}-1\right| <
      \frac{\mathit{grp\_width}}{\bar{\lambda}},
  @f]
  where @f$ \bar{\lambda} @f$ is the average @a WAVEL of the new system and 
  @f$ \bar{z}_\mathrm{old} @f$ is the average @a LINEZ of the previous one. If
  the condition is met, the two systems are regarded as independent, otherwise
  they are merged into one.
  The procedure is iterated until the whole redshift list has been scanned.
  
  If @a redsh is null, an error is issued. If no candidate system is found and 
  the verbosity level is greater than 0, a warning is issued. 
 */
/*----------------------------------------------------------------------------*/

/*
cpl_table *espda_redsh_group(cpl_table *redsh, double grp_width, 
    int grp_num, cpl_size *syst_num) {
    extern int v_level;
    extern char syst_id1, syst_id2;
    espda_util_error(redsh == NULL, CPL_ERROR_NULL_INPUT, NULL);    
    espda_struct_sort(redsh, 1, "LINEZ", FALSE);
    int limit = 0, size = cpl_table_get_nrow(redsh);
    double z_ave_old = 0.0;
    *syst_num = 0;
    espda_syst_init();        
    for (int k = 0; k < size - grp_num - 1; k++) {
        int found_flag = 0, same_flag = 0, l;
        double z_ave, null_errave, null_rms;
        for (l = grp_num; l < size - k; l++) { 
            espda_util_func(espda_spec_ave(redsh, "LINEZ", NULL, NULL, k, 
                k + l, &z_ave, &null_errave, &null_rms), SPEC_AVE, 0);

            // Distinct transitions are counted
            char line_id_old[LEN_STRING];
            int line_id_num = 0;
            cpl_table *line_id_sort = espda_struct_extract(redsh, 0, k, l);
            espda_struct_sort(line_id_sort, 1, "LINEID", FALSE);
            for (int m = 0; m < cpl_table_get_nrow(line_id_sort); m++) {
                const char *line_id = cpl_table_get_string(line_id_sort, 
                    "LINEID", m);
                if (strcmp(line_id, line_id_old)) {
                    line_id_num++;
                    strcpy(line_id_old, line_id);
                }
            }
            
            // Redshift of first and last line are computed
            double z_first = cpl_table_get_double(redsh, "LINEZ", k,     NULL); 
            double z_last  = cpl_table_get_double(redsh, "LINEZ", k + l, NULL); 
            double check_first = fabs(((1 + z_ave)     / (1 + z_first) - 1) *
                cpl_table_get_double(redsh, "WAVEL", k,     NULL));
            double check_last  = fabs(((1 + z_ave)     / (1 + z_last)  - 1) *
                cpl_table_get_double(redsh, "WAVEL", k + l, NULL));
            double check_ave   = fabs(((1 + z_ave_old) / (1 + z_ave)   - 1) *
                cpl_table_get_column_mean(line_id_sort, "WAVEL"));
            cpl_table_delete(line_id_sort);
            
            // Groups are checked w.r.t. redshifts and number of transitions
            if (check_first > grp_width || check_last > grp_width ||
                line_id_num < grp_num) {
                break;
            }
            else {
                if (check_ave < grp_width && found_flag == 0) {
                    same_flag = 1;
                }
                found_flag = 1;
                z_ave_old = z_ave;
            }
        }
        if (found_flag == 1 && k + l != limit) {

            // System ID is updated in the line list
            if (same_flag == 1) {
                espda_syst_dec();
            }
            else {
                ++*syst_num;
            }
            char syst_id[4];
            espda_syst_adv(syst_id);
            for (int m = 0; m < l; m++) {
                if (strcmp(cpl_table_get_string(redsh, "SYSTID", k + m),
                    INV_STRING) && same_flag != 1) {
                    cpl_table *row = espda_struct_extract(redsh, 0, k + m, 
                        1);
                    cpl_table_set_string(row, "SYSTID", 0, syst_id);
                    cpl_table_insert(redsh, row, cpl_table_get_nrow(redsh));
                    cpl_table_delete(row);
                }
                else {
                    cpl_table_set_string(redsh, "SYSTID", k + m, syst_id);
                }
            }
        }
        limit = k + l;
    }
    espda_syst_init();
    cpl_table *syst = NULL;
    if (*syst_num > 0) {
        espda_print_verb(1, cpl_msg_info(cpl_func, "Found %i candidate systems:",
            *syst_num));
        espda_print_verb(1, espda_print_syst(redsh, *syst_num));
        syst = espda_struct_extract(redsh, 1, "SYSTID", CPL_NOT_EQUAL_TO, 
            INV_STRING);
    }
    else { 
        espda_print_verb(1, cpl_msg_warning(cpl_func, "No candidate systems "
            "found!"));
    }
    return syst;
}
*/

/*----------------------------------------------------------------------------*/
/**
 */
/*----------------------------------------------------------------------------*/

cpl_table *espda_redsh_group(cpl_table *redsh, double mult_dz, 
    double dz, cpl_size *syst_num) {
    extern int v_level;
    extern char syst_id1, syst_id2;
    espda_util_error(redsh == NULL, CPL_ERROR_NULL_INPUT, NULL);    
    espda_struct_sort(redsh, 1, "LINEZ", FALSE);
    cpl_size size = cpl_table_get_nrow(redsh);
    *syst_num = 0;
    espda_syst_init();        
    double linez_old = -dz;
    for (int k = 1; k < size; k++) {
        double linez_check = cpl_table_get_double(redsh, "LINEZ", k, NULL) - 
            cpl_table_get_double(redsh, "LINEZ", k - 1, NULL);
        int lineid_check = strcmp(cpl_table_get_string(redsh, "LINEID", k), 
            cpl_table_get_string(redsh, "LINEID", k - 1));
        int multid_check = strcmp(cpl_table_get_string(redsh, "MULTID", k), 
            cpl_table_get_string(redsh, "MULTID", k - 1)) + 
            !strcmp(cpl_table_get_string(redsh, "MULTID", k), "");
        if (linez_check < mult_dz && lineid_check != 0 && 
            multid_check == 0) {
	    char syst_id[4];
            double group_check = cpl_table_get_double(redsh, "LINEZ", k, NULL) 
                - linez_old;
            linez_old = cpl_table_get_double(redsh, "LINEZ", k, NULL);
            if (group_check > dz) {
                ++*syst_num;
                espda_syst_adv(syst_id);
            } 
            cpl_table_set_string(redsh, "SYSTID", k - 1, syst_id);
            cpl_table_set_string(redsh, "SYSTID", k,     syst_id);
        }
    }
    espda_syst_init();
    cpl_table *syst = NULL;
    if (*syst_num > 0) {
        espda_print_verb(1,
			 cpl_msg_info(cpl_func, "Found %i candidate systems:",
				      (int)*syst_num));
        espda_print_verb(1, espda_print_syst(redsh, *syst_num));
        syst = espda_struct_extract(redsh, 1, "SYSTID", CPL_NOT_EQUAL_TO, 
            INV_STRING);
    }
    else { 
        espda_print_verb(1, cpl_msg_warning(cpl_func, "No candidate systems "
            "found!"));
    }
    return syst;
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Create a list of redshifts.
  @param   line  A list of lines (LINE_SYST).
  @param   ion   A list of ionic transitions (ION).
  @param   z_em  The quasar emission redshift.
  @return  The list of redshifts (LINE_SYST).
  
  This functions matches a list of lines @a line and a list of ionic transitions
  @a ion and creates a list with all possible redshift values for the lines 
  (i.e. not exceeding the quasar emission redshift @a z_em). The list of ionic
  transitions is trimmed beforehand to remove transitions outside the available 
  wavelength range of the instrument.

  If @a line has @f$ N @f$ rows, and @f$ M @f$ ionic transitions from @a ion are
  considered, the list of redshifts has in principle @f$ N \times M @f$ rows in
  total, computed as follows: @f[
    z_{m,n}=\frac{\lambda_n}{\lambda_m}-1,
  @f]
  where @f$ \lambda_n @f$ are taken from the @a WAVEL column in @a line, 
  @f$ \lambda_m @f$ are taken from the @a WAVEL column in @a ion, and 
  @f$ z_{m,n} @f$ are saved in the @a LINEZ column in the redshift list, 
  @f$ n @f$ spanning from 1 to @f$ N @f$ and @f$ m @f$ spanning from 1 to 
  @f$ M @f$.
  All values of @f$ z_{m,n}>\mathit{z\_em} @f$ are removed. 
  For details on the trimming of @a ion, see @p espda_redsh_trim.
  
  If @a line or @a ion are NULL, an error is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_table *espda_redsh_list(cpl_table *line, cpl_table *ion, double z_em) {
    extern int v_level;
    espda_util_error(line == NULL, CPL_ERROR_NULL_INPUT, NULL);    
    espda_util_error(ion  == NULL, CPL_ERROR_NULL_INPUT, NULL); 
    cpl_array *line_cols  = cpl_table_get_column_names(line);
    cpl_table *ion_trim   = espda_redsh_trim(ion, z_em);
    cpl_size   line_size  = cpl_table_get_nrow(line);
    cpl_size   ion_size   = cpl_table_get_nrow(ion_trim);
    cpl_size   redsh_size = line_size * ion_size;
    //cpl_table *redsh = cpl_table_new(redsh_size);
    //espda_line_syst(redsh);
    
    cpl_table *redsh = cpl_table_new(1);
    espda_line_syst(redsh);
    for (int j = 0; j < ion_size; j++) {
        cpl_table_insert(redsh, line, cpl_table_get_nrow(redsh));
    }
    cpl_table_fill_column_window_string(redsh, "SYSTID", 0, 
					redsh_size, INV_STRING);
    espda_struct_sort(redsh, 1, "WAVEL", FALSE);
   
    espda_print_verb(1, cpl_msg_info(cpl_func, "Created list of %i possible "
        "redshift values (%i lines times %i possible transitions).", 
        (int)redsh_size, (int)line_size, (int)ion_size));
    
    // Redshift list is filled
    for (int i = 0; i < line_size; i++) {
        for (int j = 0; j < ion_size; j++) {
            int k = j + i * ion_size;
            
            // Columns are initialized with the same content of the line list
	    /*
	      for (int l = 0; l < cpl_array_get_size(line_cols); l++) {
                const char *col = cpl_array_get_string(line_cols, l);
                espda_struct_copy(redsh, col, k, line, col, i);
                cpl_table_fill_column_window_string(redsh, "SYSTID", 0, 
                    redsh_size, INV_STRING);
		    }*/

            // Line ID, redshift, and group ID are updated
            double linez = cpl_table_get_double(line, "WAVEL", i, NULL) /
                cpl_table_get_double(ion_trim,  "WAVEL", j, NULL) - 1.0;
            cpl_table_set_double(redsh, "LINEZ",     k, linez);
            cpl_table_set_string(redsh, "LINEID",    k,
                cpl_table_get_string(ion_trim, "LINEID",     j));
            cpl_table_set_string(redsh, "MULTID",    k,
                cpl_table_get_string(ion_trim, "MULTID",     j));
            cpl_table_set_int   (redsh, "GRPID",     k,
                cpl_table_get_int   (line, "GRPID",     i, NULL));
        }
    }
    espda_struct_erase(redsh, 1, "LINEZ", CPL_LESS_THAN, 0.0);
    cpl_array_delete(line_cols);
    cpl_table_delete(ion_trim);
    
    // Redshift values exceeding the emission redshift are removed
    cpl_table_unselect_all(redsh);
    cpl_table_or_selected_double(redsh, "LINEZ", CPL_GREATER_THAN, z_em);
    cpl_table_erase_selected(redsh);
    return redsh;
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Trim a list of ionic transitions.
  @param   ion   A list of ionic transitions (ION).
  @param   z_em  The quasar emission redshift.
  @return  The trimmed list of ionic transitions (ION).
  
  This function takes a list of ionic transitions @a ion and remove all lines 
  that cannot fall in the usable wavelength range of the spectrum (based on the 
  quasar emission redshift @a z_em). 
  
  All transitions with @f$ \mathit{WAVEL}<\mathit{WAVEL_START}/
  (1+\mathit{z\_corr}) @f$ are removed, where @f$ \mathit{z\_corr} @f$ is a 
  maximum redshift defined from @f$ \mathit{z_em} @$ by taking into account an
  uncertainty @f$ \mathit{ZEM\_UNC}=10^3\mathrm{ km s}^{-1} @f$ in its value:
  @f[
    \mathit{z\_corr}=\frac{\mathit{z\_em}(2c+\mathit{ZEM\_UNC}) 
        + 2\mathit{ZEM\_UNC}}{2c-\mathit{ZEM\_UNC}},
  @f]
  @f$ c @f$ being the speed of light. Also, all transitions with 
  @f$ \mathit{WAVEL}>\mathit{WAVEL_END} are removed.
  
  If @a ion is NULL, an error is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_table *espda_redsh_trim(cpl_table  *ion, double z_em) {
    extern int v_level;
    espda_util_error(ion  == NULL, CPL_ERROR_NULL_INPUT, NULL);    
    double z_corr = (z_em * (2 * LIGHTSPEED + ZEM_UNC) + 2 * ZEM_UNC) /
                          (2 * LIGHTSPEED - ZEM_UNC);
    double wavel_min = WAVEL_START / (1.0 + z_corr);
    double wavel_max = WAVEL_END;
    cpl_table_unselect_all(ion);
    cpl_table_or_selected_double (ion, "WAVEL", CPL_GREATER_THAN, wavel_min);
    cpl_table_and_selected_double(ion, "WAVEL", CPL_LESS_THAN,    wavel_max);
    cpl_table *ion_trim = cpl_table_extract_selected(ion);
    return ion_trim;
}

/**@}*/

