 /* $Id: espda_split.c 172941 2015-05-05 15:32:03Z gcupani $
 *
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Author: gcupani $
 * $Date: 2015-05-05 17:32:03 +0200 (Tue, 05 May 2015) $
 * $Revision: 172941 $
 * $Name:  $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include "espda_split.h"

/*----------------------------------------------------------------------------*/
/**
  @defgroup  espda_split  Splitting
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief   Split a line group from a spectrum.
  @param   fspec_in   A full spectrum (SPEC_FULL; may be NULL).
  @param   rspec_in   A rebinned spectrum (SPEC_RMS).
  @param   line_in    A list of lines (LINE_VOIGT).
  @param   grp_sel    The group to be selected.
  @param   grp_list   The list of line groups.
  @param   fspec_out  The full spectrum of the line group (SPEC_FULL).
  @param   rspec_out  The rebinned spectrum of the line group (SPEC_RMS).
  @param   line_out   The list of lines in the group (LINE_VOIGT).
  @return  CPL error code.

  This function takes a spectrum and extracts the spectral regions corresponding
  to a group of lines @a grp_sel. A group is a set of lines that must be fitted
  together because their individual fitting ranges (as defined by columns 
  @a WAVELMIN, @a WAVELMAX) overlap with each other. The function works 
  simultaneously on a full spectrum @a fspec_in and on a rebinned spectrum 
  @a rspec_in, producing the corresponding spectra @a fspec_out, @a rspec_out 
  for the line group. Similarly, it extracts from a line list @a line_in the 
  list of lines in the group @a line_out. The lines are selected using the 
  column @a GRPID, while the spectral region are selected using the columns 
  @a WAVELMIN, @a WAVELMAX in @a line_out. A list of @a grp_list with all the 
  line groups in @a line_in is also produced. If @a grp_sel is 0, all groups are
  selected.
  
  If @a rspec_in or @a line_in are NULL, an error is issued. If @a fspec_in is 
  NULL, @a fspec_out is also NULL; in this case, if the verbosity level is 
  greater than 0, a warning is issued.
  
  N.B. This function will be possibly deprecated in a later release.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_split_group(cpl_table *fspec_in, cpl_table *rspec_in,
    cpl_table *line_in, int grp_sel, cpl_array **grp_list, 
    cpl_table **fspec_out, cpl_table **rspec_out, cpl_table **line_out) { 
    extern int v_level;
    espda_util_error(rspec_in == NULL, CPL_ERROR_NULL_INPUT, 
        cpl_error_get_code());
    espda_util_error(line_in  == NULL, CPL_ERROR_NULL_INPUT, 
        cpl_error_get_code());
    espda_util_warning(fspec_in == NULL, "full spectrum not provided.");

    // Select group(s)
    *fspec_out = NULL;
    if (grp_sel != 0) {
        espda_print_verb(1, cpl_msg_info(cpl_func, "Selecting group %i...", 
            grp_sel - 1));
        cpl_table_unselect_all(line_in);
        cpl_table_or_selected_int(line_in, "GRPID", CPL_EQUAL_TO, grp_sel - 1);
        if (cpl_table_count_selected(line_in) > 0) {
            *line_out = cpl_table_extract_selected(line_in);
            double wavel_min = cpl_table_get_double(*line_out, "WAVELMIN", 
                                   0,                               NULL);
            double wavel_max = cpl_table_get_double(*line_out, "WAVELMAX", 
                                   cpl_table_get_nrow(*line_out) - 1, NULL);
	    if (fspec_in != NULL) {
                cpl_table_unselect_all(fspec_in);
            	cpl_table_or_selected_double (fspec_in,    "WAVEL", 
            	                              CPL_GREATER_THAN, wavel_min);
            	cpl_table_and_selected_double(fspec_in,    "WAVEL", 
            	                              CPL_LESS_THAN,    wavel_max);
            	*fspec_out = cpl_table_extract_selected(fspec_in);
            }
            cpl_table_unselect_all(rspec_in);
            cpl_table_or_selected_double (rspec_in, "WAVEL", 
                                          CPL_GREATER_THAN, wavel_min);
            cpl_table_and_selected_double(rspec_in, "WAVEL", 
                                          CPL_LESS_THAN,    wavel_max);
            *rspec_out = cpl_table_extract_selected(rspec_in);
            espda_print_verb(1, cpl_msg_info(cpl_func, " %i/%i lines selected "
                "(%1.0f percent).", (int)   cpl_table_get_nrow(*line_out), 
                (int)   cpl_table_get_nrow(line_in),
                (double)cpl_table_get_nrow(*line_out) /
                (double)cpl_table_get_nrow(line_in) * 100.0));
        }
        else {
            cpl_msg_warning(cpl_func, "Chosen group is empty! Selecting all "
                            "groups..."); 
            *line_out      = cpl_table_duplicate(line_in);
            espda_struct_check(fspec_in, *fspec_out = 
                cpl_table_duplicate(fspec_in));
            *rspec_out = cpl_table_duplicate(rspec_in);
        }
    }
    else {
        espda_print_verb(1, cpl_msg_info(cpl_func, "Selecting all groups..."));
        *line_out      = cpl_table_duplicate(line_in);
        if (fspec_in != NULL) {
            *fspec_out = cpl_table_duplicate(fspec_in);
        }
        *rspec_out = cpl_table_duplicate(rspec_in);
    }

    // Create group array
    espda_print_verb(1, cpl_msg_info(cpl_func, "Creating group array..."));
    cpl_propertylist *grpid_col = cpl_propertylist_new();
    cpl_propertylist_append_bool(grpid_col, "GRPID", FALSE);
    cpl_table_sort(*line_out, grpid_col);
    cpl_propertylist_delete(grpid_col);
    cpl_size grp_size = cpl_table_get_nrow(*line_out);
    *grp_list = cpl_array_new(grp_size, CPL_TYPE_INT);
    int grp_old   = -1;
    int grp_count = 0;
    for (int g = 0; g < grp_size; g++) {
        int grp_g = cpl_table_get_int(*line_out, "GRPID", g, NULL);
        if (grp_old != grp_g) {
            cpl_array_set_int(*grp_list, grp_count, grp_g);
            grp_old = grp_g;
            grp_count++;
        }
    }
    if (grp_size > grp_count) {
        cpl_array_erase_window(*grp_list, grp_count, grp_size - grp_count);
        grp_size = grp_count;
    }
    cpl_propertylist *wavel_col = cpl_propertylist_new();
    cpl_propertylist_append_bool(wavel_col, "WAVEL", FALSE);
    cpl_table_sort(*line_out, wavel_col);
    cpl_propertylist_delete(wavel_col);
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Split a spectrum into chunks and over-chunks.
  @param   spec        A spectrum (SPEC_RMS or SPEC_FULL). 
  @param   prof        The instrument profile (INST_PROF; may be NULL).
  @param   line_chunk  The list of lines in the chunk (LINE_VOIGT; may be NULL).
  @param   grp_flag    The flag to identify if a group of lines is the last.
  @param   spec_over   The spectrum over-chunk (SPEC_RMS or SPEC_FULL).
  @param   spec_chunk  The spectrum chunk (SPEC_RMS or SPEC_FULL).
  @param   prof_chunk  The instrument profile in the spectrum chunk (INST_PROF).
  @return  CPL error code.

  This function extracts from a spectrum @a spec a chunk @a spec_chunk suitable
  for fitting the lines from a list @a line_chunk. A chunk contains all the 
  spectral regions (corresponding to one or more line groups) that must be 
  fitted together because their individual fitting ranges (as defined by columns 
  @a WAVELMIN, @a WAVELMAX) overlap with each other, or because there are some 
  constraints between their Voigt profile (as defined by columns @a LINEZCON, 
  @a COLDENCON, @a THERBCON, @a TURBBCON). If the instrument profile @a prof is 
  provided, a slightly-larger over-chunk @a spec_over is produced together with 
  the instrument profile in the chunk @a prof_chunk, to avoid border effects in 
  convolution; otherwise @a spec_over is equal to @a spec_chunk and 
  @a prof_chunk is null. 

  The function runs in two modes, depending on the external variable 
  @a grp_mode: if @a grp_mode is equal to "OVER" or @a grp_flag is equal to 
  "MAX", chunks are selected as described; otherwise, the whole spectrum across
  the lines of @a line_chunk (this is suitable for the last iteration of the
  continuum fitting algorithm).
  
  For details on the definition of the over-chunks, see @p espda_spec_conv. 

  If @a spec or @a line_chunk are NULL, an error is issued. If @a prof is NULL
  and the verbosity level is greater than 0, a warning is issued.

  N.B. This function may be expanded in a later release to include 
       @p espda_split_group. 
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_split_chunk(cpl_table *spec, cpl_table *prof,
    cpl_table *line_chunk, char *grp_flag, cpl_table **spec_over,
    cpl_table **spec_chunk, cpl_table **prof_chunk) { 
    extern int v_level;
    extern char *grp_mode;  
    espda_util_error(spec == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    espda_util_warning(line_chunk == NULL, "list of lines not provided.");
    espda_util_warning(prof       == NULL, "instrument profile not provided.");
    
    
    // When called by espda_fit_line, grp_mode is NOOVER, meaning that an 
    // overlapping fit is not computed (the region across line groups is not 
    // fitted) 

    espda_util_safe(line_chunk, espda_struct_sort(line_chunk, 1, "WAVELMIN", 
        FALSE));

    // Iteration up to the last one
    if (strcmp(grp_flag, "MAX") || strcmp(grp_mode, "OVER")) {
//    if (1 == 1) {
        cpl_size   line_size = cpl_table_get_nrow(line_chunk);
        cpl_table *over      = cpl_table_duplicate(spec);
        cpl_table_unselect_all(over);
        cpl_table_unselect_all(spec);
        if (prof != NULL) {
            cpl_table_unselect_all(prof);
        }
        for (int ir = 0; ir < line_size; ir++) {

            // Chunk definition
            double wavel_lo  = cpl_table_get_double(line_chunk, "WAVELMIN", ir,
                               NULL);
            double wavel_up  = cpl_table_get_double(line_chunk, "WAVELMAX", ir,
                               NULL);
            cpl_table_or_selected_double (spec, "WAVEL", CPL_NOT_LESS_THAN, 
                                          wavel_lo);
            cpl_table_and_selected_double(spec, "WAVEL", CPL_NOT_GREATER_THAN,
                                          wavel_up);
            if (prof != NULL) {
                cpl_table_or_selected_double (prof, "WAVEL", CPL_NOT_LESS_THAN,
                                              wavel_lo);
                cpl_table_and_selected_double(prof, "WAVEL", 
                                              CPL_NOT_GREATER_THAN, wavel_up);
                
                // Overchunk definition
                cpl_table *temp_prof = cpl_table_extract_selected(prof);

                int prof_lo = cpl_table_get_int(temp_prof, "PROFMIN", 0, NULL);
                int prof_up = cpl_table_get_int(temp_prof, "PROFMAX", 
                                   cpl_table_get_nrow(temp_prof) - 1, NULL);
                cpl_table_or_selected_window (over, prof_lo, prof_up - prof_lo);

// This is the correct version (not selecting overchunks across lines):
// it should be implemented, but now it makes the chi-square estimation go crazy
/*
                for (int p = 0; p < cpl_table_get_nrow(temp_prof); p++) {
                    int prof_lo = cpl_table_get_int(temp_prof, "PROFMIN", p, NULL);
                    int prof_up = cpl_table_get_int(temp_prof, "PROFMAX", p, NULL);
                    cpl_table_or_selected_window (over, prof_lo, prof_up - prof_lo);
                }
*/

                cpl_table_delete(temp_prof);
            } 
            else {
                cpl_table_or_selected_double (over, "WAVEL", CPL_NOT_LESS_THAN, 
                                              wavel_lo);
                cpl_table_and_selected_double(over, "WAVEL", 
                                              CPL_NOT_GREATER_THAN, wavel_up);
            }
        }
        *spec_over  = cpl_table_extract_selected(over);
        *spec_chunk = cpl_table_extract_selected(spec);
        if (prof != NULL) { 
            *prof_chunk = cpl_table_extract_selected(prof);
        }
        cpl_table_delete(over);
    }

    // Last iteration
    else {
        *spec_over  = cpl_table_duplicate(spec);
        if (prof != NULL) {
            cpl_size spec_size = cpl_table_get_nrow(prof);
            cpl_size prof_size = cpl_array_get_size(
                                 cpl_table_get_array(prof, "PROFNORM", 0));
            double wavel_lo    = cpl_table_get_double(spec, "WAVEL", 0, NULL);
            double wavel_up    = cpl_table_get_double(spec, "WAVEL", 
                                                      spec_size - 1, NULL);
            for (int i = 0; i < (int)prof_size - 1; i++) {
                if (cpl_array_has_invalid(
                    cpl_table_get_array(prof, "PROFNORM", i)))  {
                    wavel_lo = cpl_table_get_double(spec, "WAVEL", i + 1, NULL);
                }
            }
            for (int i = 0; i < (int)prof_size - 1; i++) {
                if (cpl_array_has_invalid(cpl_table_get_array(prof, "PROFNORM",
                                        spec_size - 1 - i))) {
                    wavel_up = cpl_table_get_double(spec, "WAVEL", 
                                   spec_size - 2 - i, NULL);
                }
            }
            cpl_table_unselect_all(spec);
            cpl_table_unselect_all(prof);
            cpl_table_or_selected_double (spec, "WAVEL", CPL_NOT_LESS_THAN, 
                                          wavel_lo);
            cpl_table_and_selected_double(spec, "WAVEL", CPL_NOT_GREATER_THAN,
                                          wavel_up);
            cpl_table_or_selected_double (prof, "WAVEL", CPL_NOT_LESS_THAN,
                                          wavel_lo);
            cpl_table_and_selected_double(prof, "WAVEL", CPL_NOT_GREATER_THAN,
                                          wavel_up);
            *spec_chunk = cpl_table_extract_selected(spec);
            *prof_chunk = cpl_table_extract_selected(prof);
        }
        else {
            *spec_chunk  = cpl_table_duplicate(spec);
        }
    }
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Split the blue and the red part.
  @param   table       A spectrum (SPEC_BASIC) or a list of lines (LINE_BASIC).
  @param   lya_zem     The redshifted wavelength of the Ly-alpha emission line.
  @param   edge        A velocity shift from the emission redshift.
  @param   table_blue  The blue part (SPEC_BASIC or LINE_BASIC).
  @param   table_red   The red part (SPEC_BASIC or LINE_BASIC).
  @return  CPL error code.

  This function splits a spectrum or line list @a table into its blue part 
  @a table_blue and its red part @a table red, respectively defined as the 
  regions bluewards and redwards from the redshifted wavelength @a lya_zem of
  the Ly-alpha emission line of the quasar. The @a edge parameter (espressed as 
  a velocity) can be used to create an overlap or a gap between regions.
  
  If @a table is NULL, an error is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_split_part(cpl_table *table, double lya_zem, double edge,
    cpl_table **table_blue, cpl_table **table_red) { 
    extern int v_level;
    espda_util_error(table == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    cpl_propertylist *col = cpl_propertylist_new();
    cpl_propertylist_append_bool(col, "WAVEL", FALSE);
    double lya_zem_blue = (1 + edge / LIGHTSPEED) * lya_zem + 0.01;
    cpl_table_unselect_all(table);
    cpl_table_or_selected_double(table, "WAVEL", CPL_NOT_GREATER_THAN, 
                                 lya_zem_blue);
    if (cpl_table_count_selected(table) > 0) {
        *table_blue = cpl_table_extract_selected(table);
        cpl_table_sort(*table_blue, col);
    }

    double lya_zem_red = (1 - edge / LIGHTSPEED) * lya_zem - 0.01;
    cpl_table_unselect_all(table);
    cpl_table_or_selected_double(table, "WAVEL", CPL_GREATER_THAN,
                                 lya_zem_red);
    if (cpl_table_count_selected(table) > 0) {
        *table_red  = cpl_table_extract_selected(table);
        cpl_table_sort(*table_red, col);
    }
    cpl_propertylist_delete(col);
    return cpl_error_get_code();
}

/**@}*/
