 /* $Id: espda_struct.c 270652 2022-02-25 08:58:39Z amodigli $
 *
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Author: amodigli $
 * $Date: 2022-02-25 08:58:39 +0000 (Fri, 25 Feb 2022) $
 * $Revision: 270652 $
 * $Name:  $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include "espda_struct.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup espda_struct    Operations on data structures
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief   Copy a table entry.
  @param   table_1  The source table.
  @param   col_1    The name of the source column.
  @param   pos_1    The source position.
  @param   table_2  The target table.
  @param   col_2    The name of the target column.
  @param   pos_2    The target position.
  @return  CPL error code.

  This function copies the entry at position @a pos_1 in column @a col_1 of 
  table @a table_1 to the position @a pos_2 in column @a col_2 of table 
  @a table_2. Entry values of type integer, double, and string are allowed; 
  other types are disregarded.

  If @a table_1, @a col_1, @a table_2, or @a col_2, are NULL, an error is 
  issued. If the entry has a wrong type and the verbosity level is greater than 
  0, a warning is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_struct_copy(cpl_table *table_1, const char *col_1, 
    cpl_size pos_1, cpl_table *table_2, const char *col_2, cpl_size pos_2) {
    espda_util_error(table_1 == NULL, CPL_ERROR_NULL_INPUT, 
        cpl_error_get_code());
    espda_util_error(col_1 == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    espda_util_error(table_2 == NULL, CPL_ERROR_NULL_INPUT, 
        cpl_error_get_code());
    espda_util_error(col_2 == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    cpl_type type_1 = cpl_table_get_column_type(table_1, col_1);
    cpl_type type_2 = cpl_table_get_column_type(table_2, col_2);
    if (type_1 == type_2) {
        if (type_1 == CPL_TYPE_INT) {
            cpl_table_set_int(table_1, col_1, pos_1,
                cpl_table_get_int(table_2, col_2, pos_2, NULL));
        }
        else if (type_1 == CPL_TYPE_DOUBLE) {
            cpl_table_set_double(table_1, col_1, pos_1,
                cpl_table_get_double(table_2, col_2, pos_2, NULL));
        }
        else if (type_1 == CPL_TYPE_STRING) {
            cpl_table_set_string(table_1, col_1, pos_1,
                cpl_table_get_string(table_2, col_2, pos_2));
        }
    }
    else {
        cpl_error_set_message(cpl_func, CPL_ERROR_TYPE_MISMATCH, "Columns %s "
            "and %s have different types. Unable to copy.", col_1, col_2);
    }
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Delete a structure.
  @param   obj  A structure type.
  @param   max  Number of structures to delete.
  @param   ...    The names of the structures.
  @return  CPL error code.
  
  This functions deletes a set of @a max structure of the same type @a obj. The 
  names of the structures must be provided as an argument list. Structures of 
  type CPL array, CPL table, and CPL vector are allowed; other types are 
  disregarded.

  If the structure has a wrong type and the verbosity level is greater than 0, 
  a warning is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_struct_delete(const char *obj, int max, ...) {
    va_list ap;
    va_start(ap, max);    
    if (!strcmp(obj, CPL_ARRAY)) {
        for (int i = 0; i < max; i++) {
            cpl_array *array = va_arg(ap, cpl_array *);
            cpl_array_delete(array);
        }
    }
    if (!strcmp(obj, CPL_TABLE)) {
        for (int i = 0; i < max; i++) {
            cpl_table *table = va_arg(ap, cpl_table *);
            cpl_table_delete(table);
        }
    }
    if (!strcmp(obj, CPL_VECTOR)) {
        for (int i = 0; i < max; i++) {
            cpl_vector *vector = va_arg(ap, cpl_vector *);
            cpl_vector_delete(vector);
        }
    }
    va_end(ap);
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Select and erase rows from a table.
  @param   table  A table.
  @param   max    Number of selection instructions.
  @param   ...      The selection instructions.
  @return  CPL error code.

  This functions selects rows from a table @a table according to a set of @a max 
  instructions. The selected rows are then erased.

  For details on the selection instructions, see @p espda_struct_select.
  
  If @a table is NULL, an error is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_struct_erase(cpl_table *table, int max, ...) {
    //extern int v_level;
    espda_util_error(table == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    cpl_table_unselect_all(table);
    va_list ap;
    va_start(ap, max);
    espda_struct_select(table, max, ap);
    if (cpl_table_count_selected(table) > 0) {
        cpl_table_erase_selected(table);
    }
    va_end(ap);
    cpl_table_select_all(table);
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Select and erase rows from a table.
  @param   table  A table.
  @param   max    Number of selection instructions.
  @param   ...      The selection instructions.
  @return  The extracted table.

  This functions selects rows from a table @a table according to a set of @a max 
  instructions. The selected rows are then extracted to form a new table.

  For details on the selection instructions, see @p espda_struct_select.
  
  If @a table is NULL, an error is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_table *espda_struct_extract(cpl_table *table, int max, ...) {
    //extern int v_level;
    espda_util_error(table == NULL, CPL_ERROR_NULL_INPUT, NULL);
    cpl_table_unselect_all(table);
    va_list ap;
    va_start(ap, max);
    espda_struct_select(table, max, ap);
    cpl_table *table_extr = cpl_table_extract_selected(table);
    va_end(ap);
    cpl_table_select_all(table);
    return table_extr;    
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Initialize a table column.
  @param   table  A table.
  @param   col    The name of the column.
  @param   type   The type of the column.
  @return  CPL error code.

  This function initializes column @a col of table @a table with void values. 
  Entry values of type integer, double, and string are allowed; other types are 
  disregarded. Void values are globally defined as @a INV_INT, @a INV_DOUBLE, 
  and @a INV_STRING respectively.

  If @a table or @a col are NULL, an error is issued. If the entry has a wrong 
  type and the verbosity level is greater than 0, a warning is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_struct_init(cpl_table *table, const char *col, 
    cpl_type type) {
    extern int v_level;
    espda_util_error(table == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    espda_util_error(col   == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    espda_util_warning(((type != CPL_TYPE_INT) && (type != CPL_TYPE_DOUBLE) && 
        (type != CPL_TYPE_STRING)), "wrong type provided.");
    if (cpl_table_has_column(table, col) != 1) {
        cpl_table_new_column(table, col, type); 
        cpl_size size = cpl_table_get_nrow(table); 
        if (type == CPL_TYPE_INT) { 
            cpl_table_fill_column_window_int(table, col, 0, size, INV_INT); 
        } 
        else if (type == CPL_TYPE_DOUBLE) { 
            cpl_table_fill_column_window_double(table, col, 0, size, 
                INV_DOUBLE); 
        } 
        else if (type == CPL_TYPE_STRING) { 
            cpl_table_fill_column_window_string(table, col, 0, size, 
                INV_STRING); 
        } 
    }
    return cpl_error_get_code(); 
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Select rows from a table.
  @param   table  A table.
  @param   max    Number of selection instructions.
  @param   ap     The selection instructions.
  @return  CPL error code.
  
  This functions selects rows from a table @a table according to a set of @a max 
  instructions specified by the argument list @a ap. The behavior depends on 
  @a max:

    - If @a max is 0, @a ap must be a couple of integer values @a start and 
      @a count. In this case, the window of @a count rows starting from row 
      @a start is selected.
    - If @a max is greater than 0, @a ap must be a set of @a max conditions. 
      Each condition is a triplet of values: a string @a col, a CPL table select
      operator @a op and a double @a value. All rows of @a col where @a value 
      satisfy the constraint set by @a op are selected. If more than one 
      condition is provided, each new selection acts on the unselected rows.
 
  If @a table is NULL, an error is issued.
  
  N.B. This function is not meant to be called directly, but only through 
       @p espda_struct_erase or @p espda_struct_extract.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_struct_select(cpl_table *table, int max, va_list ap) {
    espda_util_error(table == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    if (max == 0) {
        cpl_size start = va_arg(ap, cpl_size);
        cpl_size count = va_arg(ap, cpl_size);
        cpl_table_or_selected_window(table, start, count);
    }
    else {
        for (int i = 0; i < max; i++) {
            const char *col = va_arg(ap, const char *);
            cpl_type type = cpl_table_get_column_type(table, col);
            cpl_table_select_operator op = va_arg(ap, 
                cpl_table_select_operator);
            if (type == CPL_TYPE_INT) {
                int value = va_arg(ap, int);
                cpl_table_or_selected_int(table, col, op, value);
            }
            else if (type == CPL_TYPE_DOUBLE) {
                double value = va_arg(ap, double);
                cpl_table_or_selected_double(table, col, op, value);
            }
            else if (type == CPL_TYPE_STRING) {
                char *value = va_arg(ap, char *);
		if (op == CPL_EQUAL_TO) {
		    cpl_table_or_selected_string(table, col,
						 CPL_NOT_GREATER_THAN, value);
		    cpl_table_and_selected_string(table, col,
						  CPL_NOT_LESS_THAN, value);
		}
		else {  
                    cpl_table_or_selected_string(table, col, op, value);
		}
            }
        }
    }
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Sorts a table.
  @param   table  A table.
  @param   max    Number of selection instructions.
  @param   ...      The selection instructions.
  @return  CPL error code.
  
  This functions sorts a table @a table according to a set of @a max 
  instructions specified by the argument list @a ap. Each instruction is a 
  couple of values: a string @a col and an boolean @a ord. If @a ord is FALSE 
  (TRUE), @a table is sorted by column @a col in ascending (descending) order. 
  If more than one instruction is provided, each new instruction is applied only
  to the rows that have equal values in the column specified by the previous 
  instruction.
    
  If @a table is NULL, an error is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_struct_sort(cpl_table *table, int max, ...) {
    espda_util_error(table == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    cpl_propertylist *sort = cpl_propertylist_new();
    va_list ap;
    va_start(ap, max);
    for (int i = 0; i < max; i++) {
        char *col = va_arg(ap, char *);
        int   ord = va_arg(ap, int);
        cpl_propertylist_append_bool(sort, col, ord);
    }
    cpl_table_sort(table, sort);
    cpl_propertylist_delete(sort);
    va_end(ap);
    return cpl_error_get_code();
}

/**@}*/
