/* $Id: espda_mask_spec.c 288981 2024-04-25 14:36:40Z amodigli $
 *
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Author: amodigli $
 * $Date: 2024-04-25 14:36:40 +0000 (Thu, 25 Apr 2024) $
 * $Revision: 288981 $
 * $Name:  $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>

#include "espda_check.h"
#include "espda_dfs.h"
#include "espda_frame.h"
#include "espda_mask.h"
#include "espda_param.h"
#include "espda_spec.h"
#include "espda_struct.h"
#include "espda_util.h"
#include "espda_wrap.h"

/*-----------------------------------------------------------------------------
                             Plugin registration
 -----------------------------------------------------------------------------*/

int cpl_plugin_get_info(cpl_pluginlist * list);

/*-----------------------------------------------------------------------------
                            Private function prototypes
 -----------------------------------------------------------------------------*/

static int espda_mask_spec_create(cpl_plugin *);
static int espda_mask_spec_exec(cpl_plugin *);
static int espda_mask_spec_destroy(cpl_plugin *);
static int espda_mask_spec(cpl_frameset *, const cpl_parameterlist *);

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static char espda_mask_spec_description[] =
"This recipe creates a mask using a number of user-defined spectral regions.\n"
"\n"
"INPUT FRAMES              FORMAT  # EXPECTED  TAGS       NOTES\n"
"Full spectrum             SPEC    0 or 1      FSPEC_PRE\n"
"Rebinned  spectrum        SPEC    0 or 1      RSPEC_PRE  must be present if "
"FSPEC_PRE is not present\n"
"\n"
"OUTPUT FRAMES             FORMAT  # PRODUCED  TAGS       NOTES\n"
"Masked full spectrum      SPEC    0 or 1      FSPEC      same number as "
"FSPEC_PRE\n"
"Masked rebinned spectrum  SPEC    0 or 1      RSPEC      same number as "
"RSPEC_PRE\n"
"Mask                      MASK    1           MASK"
;

/*-----------------------------------------------------------------------------
                                Function code
 -----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module.
  @param    list    the plugin list
  @return   0 if everything is ok, 1 otherwise
  @note     Only this function is exported

  Create the recipe instance and make it available to the application using the
  interface.
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist *list)
{
    cpl_recipe *recipe = cpl_calloc(1, sizeof *recipe );
    cpl_plugin *plugin = &recipe->interface;

    if (cpl_plugin_init(plugin, CPL_PLUGIN_API, ESPDA_BINARY_VERSION,
                        CPL_PLUGIN_TYPE_RECIPE, "espda_mask_spec", "Create "
                        "a spectral mask and apply it to a spectrum",
                        espda_mask_spec_description, "Guido Cupani",
                        PACKAGE_BUGREPORT, espda_dfs_license(),
                        espda_mask_spec_create, espda_mask_spec_exec,
                        espda_mask_spec_destroy)) {
        cpl_msg_error(cpl_func, "Plugin initialization failed");
        (void)cpl_error_set_where(cpl_func);
        return 1;
    }

    if (cpl_pluginlist_append(list, plugin)) {
        cpl_msg_error(cpl_func, "Error adding plugin to list");
        (void)cpl_error_set_where(cpl_func);
        return 1;
    }

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options
  @param    plugin  the plugin
  @return   0 if everything is ok

  Defining the command-line/configuration parameters for the recipe.
 */
/*----------------------------------------------------------------------------*/
static int espda_mask_spec_create(cpl_plugin *plugin)
{
    cpl_recipe    *recipe;
    cpl_parameter *param;

    /* Do not create the recipe if an error code is already set */
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());
        return (int)cpl_error_get_code();
    }

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    /* Create the parameters list in the cpl_recipe object */
    recipe->parameters = cpl_parameterlist_new();
    if (recipe->parameters == NULL) {
        cpl_msg_error(cpl_func, "Parameter list allocation failed");
        cpl_ensure_code(0, (int)CPL_ERROR_ILLEGAL_OUTPUT);
    }

    /* Fill the parameters list */
    param = cpl_parameter_new_value("espda.espda_mask_spec.mask_limit",
				    CPL_TYPE_STRING,
				    "Limits of the regions to be masked [nm]. Each pair of "
            "colon-separated values defines a region. Regions must be "
            "separated by commas, with a trailing comma after the last region.",
				    "espda.espda_mask_spec",
				    "");
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "mask-limit");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_mask_spec.mask_exp",
				    CPL_TYPE_INT,
				    "Number of the exposure to be masked: "
				    "[0, total number of exposures], 0 meaning "
				    "all exposures",
				    "espda.espda_mask_spec",
				    0);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "mask-exp");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_mask_spec_exec(cpl_plugin *plugin)
{
    int             recipe_status;

    cpl_recipe     *recipe;
    cpl_errorstate  initial_errorstate = cpl_errorstate_get();

    /* Return immediately if an error code is already set */
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());
        return (int)cpl_error_get_code();
    }

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    /* Verify parameter and frame lists */
    if (recipe->parameters == NULL) {
        cpl_msg_error(cpl_func, "Recipe invoked with NULL parameter list");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }
    if (recipe->frames == NULL) {
        cpl_msg_error(cpl_func, "Recipe invoked with NULL frame set");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Invoke the recipe */
    recipe_status = espda_mask_spec(recipe->frames, recipe->parameters);

    /* Ensure DFS-compliance of the products */
    if (cpl_dfs_update_product_header(recipe->frames)) {
        if (!recipe_status) recipe_status = (int)cpl_error_get_code();
    }

    if (!cpl_errorstate_is_equal(initial_errorstate)) {
        /* Dump the error history since recipe execution start.
           At this point the recipe cannot recover from the error */
        cpl_errorstate_dump(initial_errorstate, CPL_FALSE, NULL);
    }

    return recipe_status;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_mask_spec_destroy(cpl_plugin *plugin)
{
    cpl_recipe * recipe;

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    cpl_parameterlist_delete(recipe->parameters);

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Interpret the command line options and execute the data processing
  @param    frameset   the frames list
  @param    parlist    the parameters list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_mask_spec(cpl_frameset *frameset,
			   const cpl_parameterlist *parlist) {
    extern int v_level;
    extern char *qual;

/*
 * Input is checked and loaded. The string parameter mask_limit is converted
 * into region endpoints.
 */

    cpl_msg_info(cpl_func, "*** Check and load input ***");
    /* check for invalid input */
    if(espda_files_dont_exist(frameset) != CPL_ERROR_NONE) {
    	return CPL_ERROR_BAD_FILE_FORMAT;
    }
    espda_dfs_ensure(espda_dfs_groups(frameset));
    /*
    int min = 0;
    if ((cpl_frameset_count_tags(frameset, EQ_SPEC)     == 0) &&
        (cpl_frameset_count_tags(frameset, REBIN_SPEC)  == 0)) {
        min = 1;
    }
    */
    cpl_recipeconfig *cfg = cpl_recipeconfig_new();
    char *cfg_tag = FSPEC_PRE;
    cpl_recipeconfig_set_tag  (cfg, cfg_tag, 0, 1);
    cpl_recipeconfig_set_input(cfg, cfg_tag, RSPEC_PRE, 0,   1);
    espda_wrap_check(frameset, cfg, cfg_tag);
    cpl_table *fspec = espda_frame_load(frameset, cfg, cfg_tag, -1);
    cpl_table *rspec = espda_frame_load(frameset, cfg, cfg_tag,  0);
    cpl_recipeconfig_delete(cfg);

    // Load and check parameters
    //espda_param_string("mask-limit", mask_limit,
    //    cpl_parameterlist_get_first_const(parlist));
    const char *mask_limit = (char *)malloc(sizeof(char) * LEN_STRING);
    espda_param_load(parlist, MASK_SPEC_MASK_LIMIT, &mask_limit);

/*
 * If user-defined spectral regions were provided, a mask is created. and
 * applied.
 */

    cpl_msg_info(cpl_func, "*** Create and apply the spectral mask ***");

    // Create mask
    //char *mask_repl = (char *)malloc(sizeof(char) * LEN_STRING); // No memory leaks, but crashes the recipe
    char mask_repl[LEN_STRING];
    espda_dfs_ensure(espda_param_replace(mask_limit, ":", ",", mask_repl));
    printf("%s %s\n", mask_limit, mask_repl);
    cpl_array *mask_exp_a = espda_param_extract(mask_repl, ",", 0, 2,
        CPL_TYPE_DOUBLE);
    cpl_array *mask_lim_a = espda_param_extract(mask_repl, ",", 1, 2,
        CPL_TYPE_DOUBLE);
    cpl_array_dump(mask_exp_a, 0, 0, NULL);
    cpl_vector *mask_exp = NULL, *mask_lim = NULL;
    espda_util_safe(mask_exp_a, mask_exp = cpl_vector_wrap(
        cpl_array_get_size(mask_exp_a), cpl_array_get_data_double(mask_exp_a)));
    espda_util_safe(mask_lim_a, mask_lim = cpl_vector_wrap(
        cpl_array_get_size(mask_lim_a), cpl_array_get_data_double(mask_lim_a)));
    cpl_table *mask_out = NULL;
    espda_util_func(mask_out = espda_mask_create(mask_exp, mask_lim),
        MASK_CREATE, 1);
    cpl_vector_delete(mask_exp);
    cpl_vector_delete(mask_lim);

    // Apply mask
    cpl_size fspec_nrow = 0;
    cpl_size rspec_nrow = 0;
    if (mask_out != NULL) {
        if (fspec != NULL) {
            qual = "Full    ";
            espda_util_func(espda_mask_apply(fspec, mask_out, &fspec_nrow),
                MASK_APPLY, 1);
        }
        qual = "Rebinned";
        espda_util_func(espda_mask_apply(rspec, mask_out, &rspec_nrow),
            MASK_APPLY, 1);

        if (rspec != NULL &&
            cpl_table_get_column_max(mask_out, "EXP") > 0) {
            cpl_msg_info(cpl_func, " Recomputing rebinned spectrum...");
            int    kappa = 3;
            double wavel_min = cpl_table_get_column_min(rspec, "WAVEL");
            double wavel_max = cpl_table_get_column_max(rspec, "WAVEL");
            double vel_step  = 2 * LIGHTSPEED * asinh(0.5 *
                cpl_table_get_double(rspec, "PIXSIZE", 0, NULL) /
                cpl_table_get_double(rspec, "WAVEL",   0, NULL));
            espda_util_func(rspec = espda_spec_rebin(fspec, kappa, wavel_min,
                wavel_max, vel_step), SPEC_REBIN, 0);
            cpl_size null_nrow;
            espda_util_func(espda_mask_apply(rspec, mask_out, &null_nrow),
                MASK_APPLY, 0);
        }
    }

/*
 * Output is exported. QC parameters are saved. Input spectra may have different
 * redshifts; they are consistently propagated to the products.
 */

    cpl_msg_info(cpl_func, "*** Export products into frames ***");

    // Create property lists
    cpl_propertylist *flist    = cpl_propertylist_new();
    cpl_propertylist *rlist    = cpl_propertylist_new();
    cpl_propertylist *masklist = cpl_propertylist_new();
    cpl_propertylist_append_int(flist, QC_MASK_NROW, fspec_nrow);
    cpl_propertylist_append_int(rlist, QC_MASK_NROW, rspec_nrow);

    // Create file names
    espda_frame_name(fname,    FSPEC);
    espda_frame_name(rname,    RSPEC);
    espda_frame_name(maskname, MASK);

    // Save frames
    espda_dfs_ensure(espda_frame_export(frameset, parlist, fspec,    flist,
        "espda_mask_spec", FSPEC, fname));
    espda_dfs_ensure(espda_frame_export(frameset, parlist, rspec,    rlist,
        "espda_mask_spec", RSPEC, rname));
    espda_dfs_ensure(espda_frame_export(frameset, parlist, mask_out, masklist,
        "espda_mask_spec", MASK,  maskname));

    cpl_msg_info(cpl_func, "*** Success!! ***");
    return (int)cpl_error_get_code();

}

/**@}*/
